"""Module representing the Montreal Imaging Stress Task (MIST) protocol."""
from typing import Dict, Optional, Tuple, Union

import matplotlib.pyplot as plt
from biopsykit.protocols import BaseProtocol


class MIST(BaseProtocol):
    """Class representing the Montreal Imaging Stress Task (MIST) protocol and data collected within a MIST study."""

    def __init__(
        self, name: Optional[str] = None, structure: Optional[Dict[str, Union[None, Dict[str, int]]]] = None, **kwargs
    ):
        """Class representing the Montreal Imaging Stress Task (MIST) protocol and data collected within a MIST study.

        The general structure of the MIST can be specified by passing a ``structure`` dict to the constructor.

        Up to three nested structure levels are supported:

        * 1st level: ``study part``: Different parts of the study where the MIST was conducted, such as: "Pre",
          "MIST", and "Post"
        * 2nd level: ``phase``: Different MIST phases that belong to the same *study
          part*, such as: "MIST1", "MIST2", "MIST3" (for study part "MIST") or
          "Questionnaires", "Rest", "Training" (for study part "Pre")
        * 3rd level: ``subphase``: Different MIST subphases that belong to the same *phase*, such as:
          "Baseline", "Arithmetic Task", "Feedback"


        Parameters
        ----------
        name : str, optional
            name of protocol or ``None`` to use "MIST" as default name. Default: ``None``
        structure : dict, optional
            nested dictionary specifying the structure of the MIST study.

            Up to three nested structure levels are supported:

            * 1st level: ``study part``: Different parts of the study where the MIST was conducted, such as: "Pre",
              MIST", and "Post"
            * 2nd level: ``phase``: Different MIST phases that belong to the same *study
              part*, such as: "MIST1", "MIST2", "MIST3" (for study part "MIST") or
              "Questionnaires", "Rest", "Training" (for study part "Pre")
            * 3rd level: ``subphase``: Different MIST subphases that belong to the same *phase*, such as:
              "Baseline", "Arithmetic Task", "Feedback"

            If a study part has no division into finer phases (or a phase has no division into finer subphases) the
            dictionary value can be set to ``None``.
        **kwargs
            additional parameters to be passed to ``MIST`` and its superclass, ``BaseProtocol``, such as:

            * ``saliva_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.BaseProtocol.saliva_plot`
            * ``hr_mean_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.BaseProtocol.hr_mean_plot`
            * ``hr_ensemble_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.BaseProtocol.hr_ensemble_plot`


        Examples
        --------
        >>> from biopsykit.protocols import MIST
        >>> # Example: MIST study consisting of three parts. Only the MIST part consists of different
        >>> # phases and subphases
        >>> structure = {
        >>>     "Before": None,
        >>>     "MIST": {
        >>>         "MIST1": {"BL": 60, "AT": 240, "FB": 120},
        >>>         "MIST2": {"BL": 60, "AT": 240, "FB": 120},
        >>>         "MIST3": {"BL": 60, "AT": 240, "FB": 120}
        >>>     },
        >>>     "After": None
        >>> }
        >>> MIST(name="MIST", structure=structure)


        References
        ----------
        Dedovic, K., Renwick, R., Mahani, N. K., Engert, V., Lupien, S. J., & Pruessner, J. C. (2005).
        The Montreal Imaging Stress Task: Using functional imaging to investigate the effects of perceiving and
        processing psychosocial stress in the human brain. *Journal of Psychiatry and Neuroscience*, 30(5), 319-325.

        """
        if name is None:
            name = "MIST"

        if structure is None:
            structure = {
                "Part1": None,
                "MIST": {
                    "MIST1": {"BL": 60, "AT": 240, "FB": 0},
                    "MIST2": {"BL": 60, "AT": 240, "FB": 0},
                    "MIST3": {"BL": 60, "AT": 240, "FB": 0},
                },
                "Part2": None,
            }

        test_times = kwargs.pop("test_times", [0, 30])

        hr_mean_plot_params = {"xlabel": "MIST Subphases"}
        hr_mean_plot_params.update(kwargs.pop("hr_mean_plot_params", {}))

        saliva_plot_params = {"test_title": "MIST", "xlabel": "Time relative to MIST start [min]"}
        saliva_plot_params.update(kwargs.pop("saliva_plot_params", {}))

        kwargs.update({"hr_mean_plot_params": hr_mean_plot_params, "saliva_plot_params": saliva_plot_params})
        super().__init__(name=name, structure=structure, test_times=test_times, **kwargs)

    def hr_ensemble_plot(
        self, ensemble_id: str, subphases: Optional[Dict[str, int]] = None, **kwargs
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Draw heart rate ensemble plot.

        Parameters
        ----------
        ensemble_id : str
            identifier of the ensemble data to be plotted.
            Ensemble data needs to be computed using :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_ensemble`
            first.
        subphases : dict, optional
            dictionary with phases (keys) and subphases (values - dict with subphase names and subphase durations) or
            ``None`` to retrieve MIST information from ``structure`` dict. When passing ``None``,
            it is assumed that the study part containing the MIST is named ``MIST``.
        **kwargs : dict, optional
            optional arguments for plot configuration to be passed to
            :meth:`~biopsykit.protocols.BaseProtocol.hr_ensemble_plot`.


        Returns
        -------
        fig : :class:`matplotlib.figure.Figure`
            figure object
        ax : :class:`matplotlib.axes.Axes`
            axes object


        See Also
        --------
        :meth:`~biopsykit.protocols.BaseProtocol.compute_hr_ensemble`
            compute heart rate ensemble data
        :func:`~biopsykit.protocols.plotting.hr_ensemble_plot`
            Heart rate ensemble plot

        """
        if subphases is None:
            subphases = self.structure.get("MIST", {})
        return super().hr_ensemble_plot(ensemble_id, subphases, **kwargs)



"""Module providing functions to plot data collected during sleep studies."""
import datetime
from typing import Dict, Iterable, List, Optional, Sequence, Tuple, Union

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import matplotlib.ticker as mticks
import pandas as pd
import seaborn as sns
from biopsykit.utils.datatype_helper import Acc3dDataFrame, Gyr3dDataFrame, ImuDataFrame, SleepEndpointDict
from fau_colors import colors_all

_sleep_imu_plot_params = {
    "background_color": ["#e0e0e0", "#9e9e9e"],
    "background_alpha": [0.3, 0.3],
}

_bbox_default = {
    "fc": (1, 1, 1, plt.rcParams["legend.framealpha"]),
    "ec": plt.rcParams["legend.edgecolor"],
    "boxstyle": "round",
}


def sleep_imu_plot(
    data: Union[Acc3dDataFrame, Gyr3dDataFrame, ImuDataFrame],
    datastreams: Optional[Union[str, Sequence[str]]] = None,
    sleep_endpoints: Optional[SleepEndpointDict] = None,
    downsample_factor: Optional[int] = None,
    **kwargs,
) -> Tuple[plt.Figure, Iterable[plt.Axes]]:
    """Draw plot to visualize IMU data during sleep, and, optionally, add sleep endpoints information.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        data to plot. Data must either be acceleration data (:obj:`~biopsykit.utils.datatype_helper.AccDataFrame`),
        gyroscope data (:obj:`~biopsykit.utils.datatype_helper.GyrDataFrame`), or IMU data
        (:obj:`~biopsykit.utils.datatype_helper.ImuDataFrame`).
    datastreams : str or list of str, optional
        list of datastreams indicating which type of data should be plotted or ``None`` to only plot acceleration data.
        If more than one type of datastream is specified each datastream is plotted row-wise in its own subplot.
        Default: ``None``
    sleep_endpoints : :obj:`~biopsykit.utils.datatype_helper.SleepEndpointDict`
        dictionary with sleep endpoints to add to plot or ``None`` to only plot IMU data.
    downsample_factor : int, optional
        downsample factor to apply to raw input data before plotting or ``None`` to not downsample data before
        plotting (downsample factor 1). Default: ``None``
    **kwargs
        optional arguments for plot configuration.
        To configure which type of sleep endpoint annotations to plot:

        * ``plot_sleep_onset``: whether to plot sleep onset annotations or not: Default: ``True``
        * ``plot_wake_onset``: whether to plot wake onset annotations or not: Default: ``True``
        * ``plot_bed_start``: whether to plot bed interval start annotations or not: Default: ``True``
        * ``plot_bed_end``: whether to plot bed interval end annotations or not: Default: ``True``
        * ``plot_sleep_wake``: whether to plot vspans of detected sleep/wake phases or not: Default: ``True``

        To style general plot appearance:

        * ``axs``: pre-existing axes for the plot. Otherwise, a new figure and axes objects are created and
          returned.
        * ``figsize``: tuple specifying figure dimensions
        * ``palette``: color palette to plot different axes from input data

        To style axes:

        * ``xlabel``: label of x axis. Default: "Time"
        * ``ylabel``: label of y axis. Default: "Acceleration :math:`[m/s^2]`" for acceleration data and
          "Angular Velocity :math:`[°/s]`" for gyroscope data

        To style legend:

        * ``legend_loc``: location of legend. Default: "lower left"
        * ``legend_fontsize``: font size of legend labels. Default: "smaller"


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    axs : list of :class:`~matplotlib.axes.Axes`
        list of subplot axes objects

    """
    axs: List[plt.Axes] = kwargs.pop("ax", kwargs.pop("axs", None))

    sns.set_palette(kwargs.get("palette", sns.light_palette(colors_all.fau, n_colors=4, reverse=True)[:-1]))

    if datastreams is None:
        datastreams = ["acc"]
    if isinstance(datastreams, str):
        # ensure list
        datastreams = [datastreams]

    fig, axs = _sleep_imu_plot_get_fig_axs(axs, len(datastreams), **kwargs)

    downsample_factor = _sleep_imu_plot_get_downsample_factor(downsample_factor)

    if len(datastreams) != len(axs):
        raise ValueError(
            "Number of datastreams to be plotted must match number of provided subplots! Expected {}, got {}.".format(
                len(datastreams), len(axs)
            )
        )

    for ax, ds in zip(axs, datastreams):
        _sleep_imu_plot(
            data=data,
            datastream=ds,
            downsample_factor=downsample_factor,
            sleep_endpoints=sleep_endpoints,
            ax=ax,
            **kwargs,
        )

    fig.tight_layout()
    fig.autofmt_xdate(rotation=0, ha="center")
    return fig, axs


def _sleep_imu_plot_get_fig_axs(axs: List[plt.Axes], nrows: int, **kwargs):
    figsize = kwargs.get("figsize", None)

    if isinstance(axs, plt.Axes):
        # ensure list (if only one Axes object is passed to sleep_imu_plot() instead of a list of Axes objects)
        axs = [axs]
    if axs is None:
        fig, axs = plt.subplots(figsize=figsize, nrows=nrows)
    else:
        fig = axs[0].get_figure()
    if isinstance(axs, plt.Axes):
        # ensure list (if nrows == 1 only one axes object will be created, not a list of axes)
        axs = [axs]

    return fig, axs


def _sleep_imu_plot_get_downsample_factor(downsample_factor: int):
    if downsample_factor is None:
        downsample_factor = 1
    # ensure int
    downsample_factor = int(downsample_factor)
    if downsample_factor < 1:
        raise ValueError("'downsample_factor' must be >= 1!")
    return downsample_factor


def _sleep_imu_plot(
    data: pd.DataFrame,
    datastream: str,
    downsample_factor: int,
    sleep_endpoints: SleepEndpointDict,
    ax: plt.Axes,
    **kwargs,
):
    legend_loc = kwargs.get("legend_loc", "lower left")
    legend_fontsize = kwargs.get("legend_fontsize", "smaller")
    ylabel = kwargs.get("ylabel", {"acc": "Acceleration [$m/s^2$]", "gyr": "Angular Velocity [$°/s$]"})
    xlabel = kwargs.get("xlabel", "Time")

    if isinstance(data.index, pd.DatetimeIndex):
        plt.rcParams["timezone"] = data.index.tz.zone

    data_plot = data.filter(like=datastream)[::downsample_factor]
    data_plot.plot(ax=ax)
    if sleep_endpoints is not None:
        kwargs.setdefault("ax", ax)
        _sleep_imu_plot_add_sleep_endpoints(sleep_endpoints=sleep_endpoints, **kwargs)

    if isinstance(data_plot.index, pd.DatetimeIndex):
        # TODO add axis style for non-Datetime axes
        ax.xaxis.set_major_formatter(mdates.DateFormatter("%H:%M:%S"))
        ax.xaxis.set_minor_locator(mticks.AutoMinorLocator(6))

    ax.set_ylabel(ylabel[datastream])
    ax.set_xlabel(xlabel)
    ax.legend(loc=legend_loc, fontsize=legend_fontsize, framealpha=1.0)


def _sleep_imu_plot_add_sleep_endpoints(sleep_endpoints: SleepEndpointDict, **kwargs):
    bed_start = pd.to_datetime(sleep_endpoints["bed_interval_start"])
    bed_end = pd.to_datetime(sleep_endpoints["bed_interval_end"])
    sleep_onset = pd.to_datetime(sleep_endpoints["sleep_onset"])
    wake_onset = pd.to_datetime(sleep_endpoints["wake_onset"])

    ax = kwargs.pop("ax")

    if isinstance(sleep_endpoints, dict):
        sleep_bouts = sleep_endpoints["sleep_bouts"]
        wake_bouts = sleep_endpoints["wake_bouts"]
        date = sleep_endpoints["date"]
    else:
        sleep_bouts = pd.DataFrame(sleep_endpoints["sleep_bouts"][0])
        wake_bouts = pd.DataFrame(sleep_endpoints["wake_bouts"][0])
        date = sleep_endpoints.index[0][1]

    date = pd.to_datetime(date)

    # 00:00 (12 am) vline (if present)
    if date == bed_start.normalize():
        ax.vlines(
            [date + pd.Timedelta("1d")],
            0,
            1,
            transform=ax.get_xaxis_transform(),
            linewidths=3,
            linestyles="dotted",
            colors=colors_all.tech,
            zorder=0,
        )

    _sleep_imu_plot_add_annotations(sleep_onset, wake_onset, bed_start, bed_end, sleep_bouts, wake_bouts, ax, **kwargs)

    # wear_time['end'] = wear_time.index.shift(1, freq=pd.Timedelta("15M"))
    # wear_time = wear_time[wear_time['wear'] == 0.0]
    # wear_time = wear_time.reset_index()
    #
    # handle = None
    # for idx, row in wear_time.iterrows():
    #     handle = ax.axvspan(row['index'], row['end'], color=colors.fau_color('wiso'), alpha=0.5, lw=0)
    # if handle is not None:
    #     handles['non-wear'] = handle

    ax.set_title("Sleep IMU Data: {} - {}".format(date.date(), (date + pd.Timedelta("1d")).date()))


def _sleep_imu_plot_add_annotations(
    sleep_onset: datetime.datetime,
    wake_onset: datetime.datetime,
    bed_start: datetime.datetime,
    bed_end: datetime.datetime,
    sleep_bouts,
    wake_bouts,
    ax: plt.Axes,
    **kwargs,
):
    legend_loc = "lower right"
    legend_fontsize = kwargs.get("legend_fontsize", "smaller")

    plot_sleep_onset = kwargs.get("plot_sleep_onset", True)
    plot_wake_onset = kwargs.get("plot_wake_onset", True)
    plot_bed_start = kwargs.get("plot_bed_start", True)
    plot_bed_end = kwargs.get("plot_bed_end", True)
    plot_sleep_wake = kwargs.get("plot_sleep_wake", True)

    if plot_sleep_onset:
        _sleep_imu_plot_add_sleep_onset(sleep_onset, ax, **kwargs)
    if plot_wake_onset:
        _sleep_imu_plot_add_wake_onset(wake_onset, ax, **kwargs)
    if plot_bed_start:
        _sleep_imu_plot_add_bed_start(sleep_onset, bed_start, ax, **kwargs)
    if plot_bed_end:
        _sleep_imu_plot_add_bed_end(wake_onset, bed_end, ax, **kwargs)
    if plot_sleep_wake:
        handles = _sleep_imu_plot_add_sleep_wake_bouts(sleep_bouts, wake_bouts, ax, **kwargs)
        legend = ax.legend(
            handles=list(handles.values()),
            labels=list(handles.keys()),
            loc=legend_loc,
            fontsize=legend_fontsize,
            framealpha=1.0,
        )
        ax.add_artist(legend)


def _sleep_imu_plot_add_sleep_onset(sleep_onset, ax: plt.Axes, **kwargs):
    bbox = kwargs.get("bbox", _bbox_default)

    # Sleep Onset vline
    ax.vlines(
        [sleep_onset],
        0,
        1,
        transform=ax.get_xaxis_transform(),
        linewidth=3,
        linestyles="--",
        colors=colors_all.nat,
        zorder=3,
    )

    # Sleep Onset Text + Arrow
    ax.annotate(
        "Sleep Onset",
        xy=(mdates.date2num(sleep_onset), 0.90),
        xycoords=ax.get_xaxis_transform(),
        xytext=(mdates.date2num(sleep_onset + pd.Timedelta("20min")), 0.90),
        textcoords=ax.get_xaxis_transform(),
        ha="left",
        va="center",
        bbox=bbox,
        arrowprops={
            "arrowstyle": "->",
            "lw": 2,
            "color": colors_all.nat,
            "shrinkA": 0.0,
            "shrinkB": 0.0,
        },
    )


def _sleep_imu_plot_add_wake_onset(wake_onset, ax: plt.Axes, **kwargs):
    bbox = kwargs.get("bbox", _bbox_default)
    # Wake Onset vline
    ax.vlines(
        [wake_onset],
        0,
        1,
        transform=ax.get_xaxis_transform(),
        linewidth=3,
        linestyles="--",
        colors=colors_all.nat,
        zorder=3,
    )

    # Wake Onset Text + Arrow
    ax.annotate(
        "Wake Onset",
        xy=(mdates.date2num(wake_onset), 0.90),
        xycoords=ax.get_xaxis_transform(),
        xytext=(mdates.date2num(wake_onset - pd.Timedelta("20min")), 0.90),
        textcoords=ax.get_xaxis_transform(),
        ha="right",
        va="center",
        bbox=bbox,
        arrowprops={
            "arrowstyle": "->",
            "lw": 2,
            "color": colors_all.nat,
            "shrinkA": 0.0,
            "shrinkB": 0.0,
        },
    )


def _sleep_imu_plot_add_bed_start(sleep_onset, bed_start, ax: plt.Axes, **kwargs):
    bbox = kwargs.get("bbox", _bbox_default)

    # Bed Start vline
    ax.vlines(
        [bed_start],
        0,
        1,
        transform=ax.get_xaxis_transform(),
        linewidth=3,
        linestyles="--",
        colors=colors_all.med,
        zorder=3,
    )
    # Bed Start Text + Arrow
    ax.annotate(
        "Bed Interval Start",
        xy=(mdates.date2num(bed_start), 0.80),
        xycoords=ax.get_xaxis_transform(),
        xytext=(mdates.date2num(sleep_onset + pd.Timedelta("20min")), 0.80),
        textcoords=ax.get_xaxis_transform(),
        ha="left",
        va="center",
        bbox=bbox,
        arrowprops={
            "arrowstyle": "->",
            "lw": 2,
            "color": colors_all.med,
            "shrinkA": 0.0,
            "shrinkB": 0.0,
        },
    )


def _sleep_imu_plot_add_bed_end(wake_onset, bed_end, ax: plt.Axes, **kwargs):
    bbox = kwargs.get("bbox", _bbox_default)

    # Bed End vline
    ax.vlines(
        [bed_end],
        0,
        1,
        transform=ax.get_xaxis_transform(),
        linewidth=3,
        linestyles="--",
        colors=colors_all.med,
        zorder=3,
    )
    # Bed End Text + Arrow
    ax.annotate(
        "Bed Interval End",
        xy=(mdates.date2num(bed_end), 0.80),
        xycoords=ax.get_xaxis_transform(),
        xytext=(mdates.date2num(wake_onset - pd.Timedelta("20min")), 0.80),
        textcoords=ax.get_xaxis_transform(),
        ha="right",
        va="center",
        bbox=bbox,
        arrowprops={
            "arrowstyle": "->",
            "lw": 2,
            "color": colors_all.med,
            "shrinkA": 0.0,
            "shrinkB": 0.0,
        },
    )


def _sleep_imu_plot_add_sleep_wake_bouts(
    sleep_bouts: pd.DataFrame, wake_bouts: pd.DataFrame, ax: plt.Axes, **kwargs
) -> Dict[str, plt.Artist]:
    handles = {}
    for (bout_name, bouts), bg_color, bg_alpha in zip(
        {"sleep": sleep_bouts, "wake": wake_bouts}.items(),
        kwargs.get("background_color", _sleep_imu_plot_params["background_color"]),
        kwargs.get("background_alpha", _sleep_imu_plot_params["background_alpha"]),
    ):
        handle = None
        for _, bout in bouts.iterrows():
            handle = ax.axvspan(bout["start"], bout["end"], color=bg_color, alpha=bg_alpha)

        handles[bout_name] = handle

    handles = {k: v for k, v in handles.items() if v is not None}
    return handles



"""Module containing different I/O functions to load and save sleep data."""
from pathlib import Path
from typing import Dict, Union

import pandas as pd
from biopsykit.utils._datatype_validation_helper import _assert_file_extension
from biopsykit.utils._types import path_t
from biopsykit.utils.datatype_helper import is_sleep_endpoint_dataframe, is_sleep_endpoint_dict

__all__ = ["save_sleep_endpoints"]


def save_sleep_endpoints(file_path: path_t, df_or_dict: Union[pd.DataFrame, Dict]):
    """Save sleep endpoints as csv or json file.

    Parameters
    ----------
    file_path: :class:`~pathlib.Path` or str
        file path to export
    df_or_dict : :class:`~pandas.DataFrame` or dict
         dataframe or dict with sleep endpoints to export

    Raises
    ------
    :exc:`~biopsykit.utils.exceptions.FileExtensionError`
        if ``df_or_dict`` is a dataframe and ``file_path`` is not a csv file

    """
    file_path = Path(file_path)
    if isinstance(df_or_dict, pd.DataFrame):
        _assert_file_extension(file_path, ".csv")
        is_sleep_endpoint_dataframe(df_or_dict)
        df_or_dict.to_csv(file_path)
    else:
        is_sleep_endpoint_dict(df_or_dict)
        # TODO save dict as json
        raise NotImplementedError(
            "Exporting sleep endpoint dictionary not implemented yet! Consider importing sleep endpoints as dataframe."
        )



"""Module providing functions to load different example data.

The data is either taken from the local file system in case biopsykit was installed manually or the example data is
downloaded into the local user folder.
"""
from pathlib import Path
from typing import Dict, Optional, Sequence, Tuple, Union
from urllib.request import urlretrieve

import pandas as pd
from biopsykit.io import (
    load_long_format_csv,
    load_pandas_dict_excel,
    load_questionnaire_data,
    load_subject_condition_list,
    load_time_log,
)
from biopsykit.io.ecg import load_hr_phase_dict
from biopsykit.io.eeg import MuseDataset
from biopsykit.io.nilspod import load_dataset_nilspod
from biopsykit.io.saliva import load_saliva_plate, load_saliva_wide_format
from biopsykit.io.sleep_analyzer import (
    WITHINGS_RAW_DATA_SOURCES,
    load_withings_sleep_analyzer_raw_file,
    load_withings_sleep_analyzer_raw_folder,
    load_withings_sleep_analyzer_summary,
)
from biopsykit.utils._types import path_t
from biopsykit.utils.datatype_helper import (
    HeartRatePhaseDict,
    HeartRateSubjectDataDict,
    SalivaMeanSeDataFrame,
    SalivaRawDataFrame,
    SleepEndpointDataFrame,
    SubjectConditionDataFrame,
    _SalivaMeanSeDataFrame,
    is_saliva_mean_se_dataframe,
)
from biopsykit.utils.file_handling import mkdirs
from tqdm.auto import tqdm

_EXAMPLE_DATA_PATH_LOCAL = Path(__file__).parent.parent.parent.joinpath("example_data")
_EXAMPLE_DATA_PATH_HOME = Path.home().joinpath(".biopsykit_data")
_REMOTE_DATA_PATH = "https://raw.githubusercontent.com/mad-lab-fau/BioPsyKit/main/example_data/"

__all__ = [
    "get_condition_list_example",
    "get_saliva_example_plate_format",
    "get_saliva_example",
    "get_saliva_mean_se_example",
    "get_mist_hr_example",
    "get_hr_result_sample",
    "get_hr_ensemble_sample",
    "get_hr_subject_data_dict_example",
    "get_hr_subject_data_dict_tuple_example",
    "get_ecg_processing_results_path_example",
    "get_ecg_path_example",
    "get_ecg_example",
    "get_ecg_example_02",
    "get_eeg_example",
    "get_sleep_analyzer_raw_file_unformatted",
    "get_sleep_analyzer_raw_file",
    "get_sleep_analyzer_raw_example",
    "get_sleep_analyzer_summary_example",
    "get_sleep_imu_example",
    "get_time_log_example",
    "get_questionnaire_example",
    "get_questionnaire_example_wrong_range",
    "get_stats_example",
]

# TODO add SHA256 check to assert whether remote example data was changed and should be re-downloaded.


def _is_installed_manually() -> bool:
    """Check whether biopsykit was installed manually and example data exists in the local path.

    Returns
    -------
    bool
        ``True`` if biopsykit was installed manually, ``False`` otherwise

    """
    return (_EXAMPLE_DATA_PATH_LOCAL / "__init__.py").is_file()


def _get_data(file_name: str) -> path_t:  # pragma: no cover
    if _is_installed_manually():
        return _EXAMPLE_DATA_PATH_LOCAL.joinpath(file_name)
    path = _EXAMPLE_DATA_PATH_HOME.joinpath(file_name)
    if path.exists():
        return path
    mkdirs(path.parent)
    return _fetch_from_remote(file_name, path)


def _fetch_from_remote(file_name: str, file_path: path_t) -> path_t:  # pragma: no cover
    """Download remote dataset (helper function).

    Parameters
    ----------
    file_name : str
        file name
    file_path : str
        path to file

    Returns
    -------
    :class:`~pathlib.Path`
        path to downloaded file

    """
    url = _REMOTE_DATA_PATH + file_name
    print(f"Downloading file {file_name} from remote URL: {url}.")
    with tqdm(unit="B", unit_scale=True, unit_divisor=1024, miniters=1, desc=file_name) as t:
        urlretrieve(url, filename=file_path, reporthook=_tqdm_hook(t))
    return file_path


def _tqdm_hook(t):  # pragma: no cover
    """Wrap tqdm instance."""
    last_b = [0]

    def update_to(b=1, bsize=1, tsize=None):
        if tsize is not None:
            t.total = tsize
        t.update((b - last_b[0]) * bsize)
        last_b[0] = b

    return update_to


def get_condition_list_example() -> SubjectConditionDataFrame:
    """Return example data for subject condition assignment.

    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.SubjectConditionDataFrame`
        dataframe with example subject condition assignment

    """
    return load_subject_condition_list(
        _get_data("condition_list.csv"), subject_col="subject", condition_col="condition"
    )


def get_saliva_example_plate_format(
    sample_id_col: Optional[str] = None,
    data_col: Optional[str] = None,
    id_col_names: Optional[Sequence[str]] = None,
    regex_str: Optional[str] = None,
    sample_times: Optional[Sequence[int]] = None,
    condition_list: Optional[Union[Sequence, Dict[str, Sequence], pd.Index]] = None,
) -> pd.DataFrame:
    r"""Return example saliva data from "plate" format.

    Parameters
    ----------
    sample_id_col: str, optional
        column name of the Excel sheet containing the sample ID. Default: "sample ID"
    data_col: str, optional
        column name of the Excel sheet containing saliva data to be analyzed.
        Default: Select default column name based on ``biomarker_type``, e.g. ``cortisol`` => ``cortisol (nmol/l)``
    id_col_names: list of str, optional
        names of the extracted ID column names. ``None`` to use the default column names (['subject', 'day', 'sample'])
    regex_str: str, optional
        regular expression to extract subject ID, day ID and sample ID from the sample identifier.
        ``None`` to use default regex string (``r"(Vp\d+) (S\d)"``)
    sample_times: list of int, optional
        times at which saliva samples were collected
    condition_list: 1d-array, optional
        list of conditions which subjects were assigned to

    Returns
    -------
    data : :class:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`
        saliva data in `SalivaRawDataFrame` format

    See Also
    --------
    :func:`~biopsykit.io.saliva.load_saliva_plate`
        loader function for saliva data in plate format

    """
    return load_saliva_plate(
        _get_data("cortisol_sample_plate.xlsx"),
        "cortisol",
        sample_id_col=sample_id_col,
        data_col=data_col,
        id_col_names=id_col_names,
        regex_str=regex_str,
        sample_times=sample_times,
        condition_list=condition_list,
    )


def get_saliva_example(sample_times: Optional[Sequence[int]] = None) -> SalivaRawDataFrame:
    """Return saliva example data.

    Parameters
    ----------
    sample_times : list of int, optional
        sample times of saliva samples in minutes

    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.SalivaRawDataFrame`
        dataframe with example raw saliva data

    """
    return load_saliva_wide_format(
        _get_data("cortisol_sample.csv"),
        saliva_type="cortisol",
        condition_col="condition",
        sample_times=sample_times,
    )


def get_saliva_mean_se_example() -> Dict[str, SalivaMeanSeDataFrame]:
    """Return dictionary with mean and standard error from example data for different saliva types.

    Returns
    -------
    dict
        dictionary with :obj:`~biopsykit.utils.datatype_helper.SalivaMeanSeDataFrame` from different saliva types

    """
    data_dict = pd.read_excel(_get_data("saliva_sample_mean_se.xlsx"), sheet_name=None)
    for key in data_dict:
        data_dict[key] = _SalivaMeanSeDataFrame(data_dict[key].set_index(["condition", "sample", "time"]))
        is_saliva_mean_se_dataframe(data_dict[key])
    return data_dict


def get_hr_subject_data_dict_example() -> HeartRateSubjectDataDict:
    """Return heart rate example data in the form of a :obj:`~biopsykit.utils.datatype_helper.HeartRateSubjectDataDict`.

    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.HeartRateSubjectDataDict`
        dictionary with heart rate time-series data from multiple subjects, each containing data from different phases.

    """
    study_data_dict_hr = {}
    subject_ids = ["Vp01", "Vp02"]
    for subject_id in subject_ids:
        file_path = _get_data(f"ecg_results/hr_result_{subject_id}.xlsx")
        study_data_dict_hr[subject_id] = pd.read_excel(file_path, sheet_name=None, index_col="time")
    return study_data_dict_hr


def get_hr_subject_data_dict_tuple_example() -> HeartRateSubjectDataDict:
    """Return HR example data as a ``HeartRateSubjectDataDict``, but with tuples as keys instead of strings.

    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.HeartRateSubjectDataDict`
        dictionary with heart rate time-series, each containing data from different phases.

    """
    study_data_dict_hr = {}
    subject_ids = ["Vp01", "Vp02"]
    keys = [("Vp01", "Control"), ("Vp01", "Intervention")]
    for subject_id, key in zip(subject_ids, keys):
        file_path = _get_data(f"ecg_results/hr_result_{subject_id}.xlsx")
        study_data_dict_hr[key] = pd.read_excel(file_path, sheet_name=None, index_col="time")
    return study_data_dict_hr


def get_hr_result_sample() -> pd.DataFrame:
    """Return heart rate results example data.

    The heart rate results example data consists of the mean normalized heart rate for different subjects,
    different study phases, and study subphases.

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with heart rate results example data

    """
    return load_long_format_csv(_get_data("hr_result_sample.csv"))


def get_hr_ensemble_sample() -> Dict[str, pd.DataFrame]:
    """Return heart rate ensemble example data.

    The example data consists of time-series heart rate of multiple subjects for different study phases,
    each synchronized and resampled to 1 Hz, and normalized to baseline heart rate.

    Returns
    -------
    dict
        dictionary with pandas dataframes containing heart rate ensemble data

    """
    return load_pandas_dict_excel(_get_data("hr_ensemble_sample_normalized.xlsx"))


def get_mist_hr_example() -> HeartRatePhaseDict:
    """Return heart rate time-series example data collected during MIST from one subject.

    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.HeartRatePhaseDict`
        dictionary with heart rate time-series data from one subject during multiple phases

    """
    return load_hr_phase_dict(_get_data("hr_sample_mist.xlsx"))


def get_ecg_path_example() -> path_t:
    """Return folder path to ECG example data.

    Returns
    -------
    :class:`~pathlib.Path` or str
        path to folder with ECG raw files

    """
    subject_ids = ["Vp01", "Vp02"]
    file_path = None
    # ensure that folder exists and data in folder is available
    for subject_id in subject_ids:
        file_path = _get_data(f"ecg/ecg_sample_{subject_id}.bin")
    return file_path.parent


def get_ecg_processing_results_path_example() -> path_t:
    """Return folder path to ECG processing results.

    Returns
    -------
    :class:`~pathlib.Path` or str
        path to folder with ECG processing results

    """
    subject_ids = ["Vp01", "Vp02"]
    file_path = None
    # ensure that folder exists and data in folder is available
    for subject_id in subject_ids:
        for file_type in ["hr_result", "rpeaks_result"]:
            file_path = _get_data(f"ecg_processing_results/{file_type}_{subject_id}.xlsx")
    return file_path.parent


def get_ecg_example() -> Tuple[pd.DataFrame, float]:
    """Return raw ECG example data from one subject.

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with raw ECG data
    sampling_rate : float
        sampling rate of recorded data

    """
    return load_dataset_nilspod(file_path=_get_data("ecg/ecg_sample_Vp01.bin"), datastreams=["ecg"])


def get_ecg_example_02() -> Tuple[pd.DataFrame, float]:
    """Return second raw ECG example data from another subject.

    Returns
    -------
    :class:`~pandas.DataFrame`
        dataframe with raw ECG data
    sampling_rate : float
        sampling rate of recorded data

    """
    return load_dataset_nilspod(file_path=_get_data("ecg/ecg_sample_Vp02.bin"), datastreams=["ecg"])


def get_sleep_analyzer_raw_file_unformatted(data_source: str) -> pd.DataFrame:
    """Return *unformatted* Withings Sleep Analyzer raw data example file.

    Parameters
    ----------
    data_source : str
        Withings Sleep Analyzer data source name.
        Must be one of ['heart_rate', 'respiration_rate', 'sleep_state', 'snoring'].

    Returns
    -------
    :class:`~pandas.DataFrame`
        Dataframe with unformatted example raw data

    """
    if data_source not in WITHINGS_RAW_DATA_SOURCES.values():
        raise ValueError(
            "Unsupported data source {}! Must be one of {}.".format(
                data_source, list(WITHINGS_RAW_DATA_SOURCES.values())
            )
        )
    ds_name = list(WITHINGS_RAW_DATA_SOURCES.keys())[list(WITHINGS_RAW_DATA_SOURCES.values()).index(data_source)]
    return pd.read_csv(_get_data(f"sleep/raw_sleep-monitor_{ds_name}.csv"))


def get_sleep_analyzer_raw_file(
    data_source: str,
    split_into_nights: Optional[bool] = True,
) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
    """Return Withings Sleep Analyzer raw data example file.

    Parameters
    ----------
    data_source : str
        Withings Sleep Analyzer data source name.
        Must be one of ['heart_rate', 'respiration_rate', 'sleep_state', 'snoring'].
    split_into_nights : bool, optional
        whether to split the dataframe into the different recording nights (and return a dictionary of dataframes)
        or not.
        Default: ``True``

    Returns
    -------
    :class:`~pandas.DataFrame` or dict of such
        dataframe (or dict of dataframes, if ``split_into_nights`` is ``True``) with Sleep Analyzer data

    """
    if data_source not in WITHINGS_RAW_DATA_SOURCES.values():
        raise ValueError(
            "Unsupported data source {}! Must be one of {}.".format(
                data_source, list(WITHINGS_RAW_DATA_SOURCES.values())
            )
        )

    ds_name = list(WITHINGS_RAW_DATA_SOURCES.keys())[list(WITHINGS_RAW_DATA_SOURCES.values()).index(data_source)]
    return load_withings_sleep_analyzer_raw_file(
        _get_data(f"sleep/raw_sleep-monitor_{ds_name}.csv"),
        data_source=data_source,
        split_into_nights=split_into_nights,
    )


def get_sleep_analyzer_raw_example(
    split_into_nights: Optional[bool] = True,
) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
    """Return Withings Sleep Analyzer example raw data.

    Parameters
    ----------
    split_into_nights : bool, optional
        ``True`` to split data into single dataframes per recording night, and return a dict of dataframes,
        ``False`` to keep all data in one dataframe.
        Default: ``True``

    Returns
    -------
    :class:`~pandas.DataFrame` or dict
        dataframe with raw sleep analyzer data or a dict of such if ``split_into_nights`` is ``True``

    """
    # ensure that all files are available
    file_list = [
        "raw_sleep-monitor_hr.csv",
        "raw_sleep-monitor_respiratory-rate.csv",
        "raw_sleep-monitor_sleep-state.csv",
        "raw_sleep-monitor_snoring.csv",
    ]
    file_path = None
    for file in file_list:
        file_path = _get_data(f"sleep/{file}")
    # get parent directory
    file_path = file_path.parent
    return load_withings_sleep_analyzer_raw_folder(file_path, split_into_nights=split_into_nights)


def get_sleep_analyzer_summary_example() -> SleepEndpointDataFrame:
    """Return Withings Sleep Analyzer example summary data.

    Returns
    -------
    :obj:`~biopsykit.utils.datatype_helper.SleepEndpointDataFrame`
        dataframe with example sleep endpoints computed from Withings Sleep Analyzer Summary data

    """
    return load_withings_sleep_analyzer_summary(_get_data("sleep/sleep.csv"))


def get_sleep_imu_example() -> Tuple[pd.DataFrame, float]:
    """Return raw IMU example data collected from a wrist-worn IMU sensor during night.

    Returns
    -------
    data : :class:`~pandas.DataFrame`
        dataframe with raw IMU data
    sampling_rate : float
        sampling rate of recorded data

    """
    return load_dataset_nilspod(file_path=_get_data("sleep_imu/sleep_imu_sample_01.bin"))


def get_eeg_example() -> MuseDataset:
    """Return raw EEG example data collected from a Muse EEG headband.

    Returns
    -------
    data : :class:`~pandas.DataFrame`
        dataframe with raw EEG data
    sampling_rate : float
        sampling rate of recorded data

    """
    return MuseDataset.from_csv_file(_get_data("eeg_muse_example.csv"), tz="Europe/Berlin")


def get_time_log_example() -> pd.DataFrame:
    """Return time log example data.

    Returns
    -------
    data : :class:`~pandas.DataFrame`
        dataframe with example time log information. The time log match the data from the two ECG data example
        functions :func:`~biopsykit.example_data.get_ecg_example` and :func:`~biopsykit.example_data.get_ecg_example_02`

    """
    return load_time_log(_get_data("ecg_time_log.xlsx"))


def get_questionnaire_example_wrong_range() -> pd.DataFrame:
    """Return questionnaire example data with score in the wrong range.

    In this example the items of the "PSS" questionnaire are coded in the wrong range ([1, 5] instead of [0, 4])
    originally defined in the paper.

    This example data is used to demonstrate BioPsyKit's feature of asserting that questionnaire score items are
    provided in the correct score range according to the original definition of the questionnaire.


    Returns
    -------
    data : :class:`~pandas.DataFrame`
        dataframe with questionnaire example data where the items of the PSS questionnaire are coded in the wrong range

    """
    return load_questionnaire_data(_get_data("questionnaire_sample_wrong_range.csv"))


def get_questionnaire_example() -> pd.DataFrame:
    """Return questionnaire example data.

    Returns
    -------
    data : :class:`~pandas.DataFrame`
        dataframe with questionnaire example data

    """
    return load_questionnaire_data(_get_data("questionnaire_sample.csv"))


def get_stats_example() -> pd.DataFrame:
    """Return example data for statistical analysis.

    Returns
    -------
    data : :class:`~pandas.DataFrame`
        dataframe with example data that can be used for statistical analysis

    """
    return load_long_format_csv(_get_data("stats_sample.csv"))



"""Module representing the Cold Face Test (CFT) protocol."""
import datetime
import warnings
from typing import Any, Dict, Optional, Tuple, Union

import matplotlib.patches as mpatch
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from biopsykit.protocols import BaseProtocol
from biopsykit.signals.ecg.plotting import hr_plot
from biopsykit.utils.exceptions import FeatureComputationError
from fau_colors import colors_all


class CFT(BaseProtocol):
    """Class representing the Cold Face Test (CFT) and data collected while conducting the CFT."""

    def __init__(
        self,
        name: Optional[str] = None,
        structure: Optional[Dict[str, int]] = None,
        **kwargs,
    ):
        """Class representing the Cold Face Test (CFT) and data collected while conducting the CFT.

        The typical structure of the CFT consists of three phases:

          * "Baseline": Time at rest before applying cold face stimulus
          * "CFT": Application of cold face stimulus
          * "Recovery": Time at rest after applying cold face stimulus


        Parameters
        ----------
        name : str
            name of protocol or ``None`` to use "CFT" as default name. Default: ``None``
        structure : dict, optional
            nested dictionary specifying the structure of the CFT.

            The typical structure of the CFT consists of three phases:

            * "Baseline": Time at rest before applying cold face stimulus
            * "CFT": Application of cold face stimulus
            * "Recovery": Time at rest after applying cold face stimulus

            The duration of each phase is specified in seconds.
            Typical durations are: 60s for *Baseline*, 120s for *CFT*, 60s for *Recovery*

            The start and duration of the CFT Exposure (``cft_start`` and ``cft_duration``) will be automatically
            extracted from the structure dictionary.
        **kwargs :
            additional parameters to be passed to ``CFT`` and its superclass, ``BaseProcessor``, such as:

            * ``cft_plot_params``: dictionary with parameters to style
              :meth:`~biopsykit.protocols.CFT.cft_plot`


        Examples
        --------
        >>> from biopsykit.protocols import CFT
        >>> # Example: CFT procedure consisting of three parts.
        >>>
        >>> structure = {
        >>>     "Baseline": 60,
        >>>     "CFT": 120,
        >>>     "Recovery": 60
        >>> }
        >>> CFT(name="CFT", structure=structure)

        """
        if name is None:
            name = "CFT"

        if structure is None:
            structure = {"Baseline": 60, "CFT": 120, "Recovery": 60}

        cft_start = structure.setdefault("Baseline", 0)
        if cft_start < 0:
            raise ValueError("'Baseline' duration must be non-negative!")

        cft_duration = structure.setdefault("CFT", 120)
        if cft_duration <= 0:
            raise ValueError("'CFT' duration must be positive!")

        self.cft_start: int = cft_start
        """Start of Cold Face Exposure in seconds. It is assumed that the time between the *beginning* of the
        recording and the *start* of the Cold Face Exposure is the Baseline. Default: 60 seconds"""

        self.cft_duration: int = cft_duration
        """Duration of the Cold Face Exposure in seconds. Default: 120 seconds"""

        cft_plot_params = {
            "background_color": None,
            "background_base_color": "#e0e0e0",
            "background_alpha": 0.2,
            "phase_names": list(structure.keys()),
        }
        cft_plot_params.update(kwargs.get("cft_plot_params", {}))
        self.cft_plot_params: Dict[str, Any] = cft_plot_params

        super().__init__(name, structure, **kwargs)

    def compute_cft_parameter(
        self,
        data: pd.DataFrame,
        index: Optional[str] = None,
        return_dict: Optional[bool] = False,
    ) -> Union[Dict, pd.DataFrame]:
        """Compute CFT parameter.

        This function computes the following CFT parameter and returns the result in a dataframe
        (or, optionally, as dictionary):

        * Baseline Heart Rate (see :meth:`~biopsykit.protocols.CFT.baseline_hr` for further information)
        * CFT Onset (see :meth:`~biopsykit.protocols.CFT.onset` for further information)
        * Peak Bradycardia (see :meth:`~biopsykit.protocols.CFT.peak_bradycardia` for further information)
        * Mean Bradycardia (see :meth:`~biopsykit.protocols.CFT.mean_bradycardia` for further information)
        * Polynomial Fit on CFT Reaction (see :meth:`~biopsykit.protocols.CFT.poly_fit` for further information)


        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data
        index : str, optional
            index value of resulting dataframe. Not needed if dictionary should be returned.
        return_dict : bool, optional
            ``True`` to return a dictionary with CFT parameters, ``False`` to return a dataframe. Default: ``False``


        Returns
        -------
        :class:`~pandas.DataFrame` or dict
            dataframe or dictionary with CFT parameter

        """
        if isinstance(index, str):
            index = [index]

        dict_cft = {}
        cft_parameter = ["onset", "peak_bradycardia", "mean_bradycardia", "poly_fit"]

        df_cft = self.extract_cft_interval(data)

        hr_baseline = self.baseline_hr(data)
        dict_cft["baseline_hr"] = hr_baseline
        dict_cft["cft_start_idx"] = data.index.get_loc(df_cft.index[0])
        for parameter in cft_parameter:
            dict_cft.update(
                getattr(self, parameter)(
                    data=df_cft,
                    is_cft_interval=True,
                    compute_baseline=False,
                    hr_baseline=hr_baseline,
                )
            )

        if return_dict:
            return dict_cft
        return pd.DataFrame([dict_cft], index=index)

    def baseline_hr(self, data: pd.DataFrame) -> float:
        """Compute mean heart rate during Baseline Interval.

        The Baseline Interval is data in the interval [``0``, ``cft_start``].

        .. warning::
            If ``cft_start`` is 0, it is assumed that no Baseline is present and the first heart rate value
            in the dataframe is used as CFT Baseline.


        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data


        Returns
        -------
        float
            mean heart rate during Baseline Interval


        Raises
        ------
        :ex:`~biopsykit.utils.exceptions.FeatureComputationError`
            if data is shorter than the expected duration of the Baseline interval

        """
        # start of baseline = start of recording
        bl_start = data.index[0]
        if self.cft_start == 0:
            warnings.warn(
                "cft_start is 0, no baseline can be extracted! "
                "Using the first heart rate value in the dataframe as baseline."
            )
        # end of baseline = start of CFT
        if isinstance(data.index, pd.DatetimeIndex):
            bl_end = bl_start + pd.Timedelta(seconds=self.cft_start)
            if bl_end > data.index[-1]:
                raise FeatureComputationError(
                    "Error computing Baseline heart rate! "
                    "The provided data is shorter than the expected Baseline interval."
                )
            # heart rate during Baseline
            hr_baseline = data.between_time(bl_start.time(), bl_end.time())
        else:
            bl_end = bl_start + self.cft_start
            if bl_end > data.index[-1]:
                raise FeatureComputationError(
                    "Error computing Baseline heart rate! The provided data is shorter than the Baseline interval."
                )
            hr_baseline = data.loc[bl_start:bl_end]
        return float(hr_baseline.mean())

    def extract_cft_interval(self, data: pd.DataFrame) -> pd.DataFrame:
        """Extract interval during which CFT was applied.

        This function extracts only the part of the data during the "actual" Cold Face Test, i.e.,
        the time during which the cold face stimulus was applied.


        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data


        Returns
        -------
        :class:`~pandas.DataFrame`
            data during application of cold face stimulus

        """
        if isinstance(data.index, pd.DatetimeIndex):
            cft_start = data.index[0] + pd.Timedelta(seconds=self.cft_start)
            cft_end = data.index[0] + pd.Timedelta(seconds=self.cft_start + self.cft_duration)
            return data.between_time(cft_start.time(), cft_end.time())
        return data.loc[self.cft_start : self.cft_start + self.cft_duration]

    def onset(
        self,
        data: pd.DataFrame,
        is_cft_interval: Optional[bool] = False,
        compute_baseline: Optional[bool] = True,
        hr_baseline: Optional[float] = None,
    ) -> Dict[str, Any]:
        """Compute **CFT onset**.

        The CFT onset is defined as the time point after beginning of the CFT Interval where three consecutive
        heart beats are lower than the Baseline heart rate (typically the Interval directly before the CFT).

        This function computes the following CFT onset parameter:

        * ``onset``: location of CFT onset. This value is the same datatype as the index of ``data``
          (i.e., either a absolute datetime timestamp or a relative timestamp in time since recording).
        * ``onset_latency``: CFT onset latency, i.e., the duration between beginning of the CFT Interval and
          CFT onset in seconds.
        * ``onset_idx``: location of CFT onset as array index
        * ``onset_hr``: heart rate at CFT onset in bpm
        * ``onset_hr_brady_percent``: bradycardia at CFT onset, i.e., relative change of CFT onset heart rate compared
          to Baseline heart rate in percent.
        * ``onset_slope``: Slope between Baseline heart rate and CFT onset heart rate, computed as:
          `onset_slope = (onset_hr - baseline_hr) / onset_latency`


        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data
        is_cft_interval : bool, optional
            ``True`` if the heart rate data passed via ``data`` contains only the CFT Interval,
            ``False`` if it contains the data during the whole CFT procedure. Default: ``False``
        compute_baseline : bool, optional
            ``True`` if Baseline Interval is included in data passed via ``data`` and Baseline heart rate
            should be computed or ``False`` if Baseline heart rate is passed separately via ``hr_baseline``.
            Default: ``True``
        hr_baseline : float, optional
            mean heart rate during Baseline Interval or ``None`` if Baseline interval is present in ``data`` and
            Baseline heart rate is computed from there. Default: ``None``


        Returns
        -------
        dict
            dictionary with CFT onset parameter

        """
        df_hr_cft, hr_baseline = self._sanitize_cft_input(data, is_cft_interval, compute_baseline, hr_baseline)

        # bradycardia mask (True where heart rate is below baseline, False otherwise)
        hr_brady = df_hr_cft < hr_baseline
        # bradycardia borders (1 where we have a change between lower and higher heart rate)
        brady_border = np.abs(np.ediff1d(hr_brady.astype(int), to_begin=0))
        # filter out the phases where we have at least 3 heart rate values lower than baseline
        brady_phases = hr_brady.groupby([np.cumsum(brady_border)]).filter(lambda df: df.sum() >= 3)
        # CFT onset is the third beat
        onset = brady_phases.index[2]
        # TODO check index handling again...
        onset_latency = (onset - df_hr_cft.index[0]).total_seconds()
        onset_idx = df_hr_cft.index.get_loc(onset)

        # heart rate at onset point
        hr_onset = np.squeeze(df_hr_cft.loc[onset])
        return {
            "onset": onset,
            "onset_latency": onset_latency,
            "onset_idx": onset_idx,
            "onset_hr": hr_onset,
            "onset_hr_percent": (1 - hr_onset / hr_baseline) * 100,
            "onset_slope": (hr_onset - hr_baseline) / onset_latency,
        }

    def peak_bradycardia(
        self,
        data: pd.DataFrame,
        is_cft_interval: Optional[bool] = False,
        compute_baseline: Optional[bool] = True,
        hr_baseline: Optional[float] = None,
    ) -> Dict[str, Any]:
        """Compute **CFT peak bradycardia**.

        The CFT peak bradycardia is defined as the maximum bradycardia (i.e., the minimum heart rate)
        during the CFT Interval.

        This function computes the following CFT peak bradycardia parameter:

        * ``peak_brady``: location of CFT peak bradycardia. This value is the same datatype as the index of ``data``
          (i.e., either a absolute datetime timestamp or a relative timestamp in time since recording).
        * ``peak_brady_latency``: CFT peak bradycardia latency, i.e., the duration between beginning of the
          CFT Interval and CFT peak bradycardia in seconds.
        * ``peak_brady_idx``: location of CFT peak bradycardia as array index
        * ``peak_brady_bpm``: CFT peak bradycardia in bpm
        * ``peak_brady_percent``: Relative change of CFT peak bradycardia heart rate compared to Baseline
          heart rate in percent.
        * ``peak_brady_slope``: Slope between Baseline heart rate and CFT peak bradycardia heart rate, computed as:
          ``peak_brady_slope = (peak_brady_bpm - baseline_hr) / peak_brady_latency``


        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data
        is_cft_interval : bool, optional
            ``True`` if the heart rate data passed via ``data`` contains only the CFT Interval,
            ``False`` if it contains the data during the whole CFT procedure. Default: ``False``
        compute_baseline : bool, optional
            ``True`` if Baseline Interval is included in data passed via ``data`` and Baseline heart rate
            should be computed or ``False`` if Baseline heart rate is passed separately via ``hr_baseline``.
            Default: ``True``
        hr_baseline : float, optional
            mean heart rate during Baseline Interval or ``None`` if Baseline interval is present in ``data`` and
            Baseline heart rate is computed from there. Default: ``None``


        Returns
        -------
        dict
            dictionary with CFT peak bradycardia parameter

        """
        df_hr_cft, hr_baseline = self._sanitize_cft_input(data, is_cft_interval, compute_baseline, hr_baseline)

        peak_brady = np.squeeze(df_hr_cft.idxmin())
        peak_brady_latency = (peak_brady - df_hr_cft.index[0]).total_seconds()
        peak_brady_idx = df_hr_cft.index.get_loc(peak_brady)
        hr_brady = np.squeeze(df_hr_cft.loc[peak_brady])
        return {
            "peak_brady": peak_brady,
            "peak_brady_latency": peak_brady_latency,
            "peak_brady_idx": peak_brady_idx,
            "peak_brady_bpm": hr_brady - hr_baseline,
            "peak_brady_percent": (hr_brady / hr_baseline - 1) * 100,
            "peak_brady_slope": (hr_brady - hr_baseline) / peak_brady_latency,
        }

    def mean_bradycardia(
        self,
        data: pd.DataFrame,
        is_cft_interval: Optional[bool] = False,
        compute_baseline: Optional[bool] = True,
        hr_baseline: Optional[float] = None,
    ) -> Dict[str, Any]:
        """Compute **CFT mean bradycardia**.

        The CFT mean bradycardia is defined as the mean bradycardia (i.e., the mean decrease of heart rate)
        during the CFT Interval.

        This function computes the following CFT mean bradycardia parameter:

        * ``mean_hr_bpm``: average heart rate during CFT Interval in bpm
        * ``mean_brady_bpm``: average bradycardia during CFT Interval, computed as:
          ``mean_brady_bpm = mean_hr_bpm - hr_baseline``
        * ``mean_brady_percent``: relative change of CFT mean bradycardia heart rate compared to Baseline
          heart rate in percent


        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data
        is_cft_interval : bool, optional
            ``True`` if the heart rate data passed via ``data`` contains only the CFT Interval,
            ``False`` if it contains the data during the whole CFT procedure. Default: ``False``
        compute_baseline : bool, optional
            ``True`` if Baseline Interval is included in data passed via ``data`` and Baseline heart rate
            should be computed or ``False`` if Baseline heart rate is passed separately via ``hr_baseline``.
            Default: ``True``
        hr_baseline : float, optional
            mean heart rate during Baseline Interval or ``None`` if Baseline interval is present in ``data`` and
            Baseline heart rate is computed from there. Default: ``None``


        Returns
        -------
        dict
            dictionary with CFT mean bradycardia parameter

        """
        df_hr_cft, hr_baseline = self._sanitize_cft_input(data, is_cft_interval, compute_baseline, hr_baseline)

        hr_mean = np.squeeze(df_hr_cft.mean())

        return {
            "mean_hr_bpm": hr_mean,
            "mean_brady_bpm": hr_mean - hr_baseline,
            "mean_brady_percent": (hr_mean / hr_baseline - 1) * 100,
        }

    def poly_fit(
        self,
        data: pd.DataFrame,
        is_cft_interval: Optional[bool] = False,
        compute_baseline: Optional[bool] = True,
        hr_baseline: Optional[float] = None,
    ) -> Dict[str, Any]:
        """Compute **CFT polynomial fit**.

        The CFT polynomial fit is computed by applying a 2nd order least-squares polynomial fit to the heart rate
        during the CFT Interval because the CFT-induced bradycardia and the following recovery is assumed to follow
        a polynomial function.

        This function computes the following CFT polynomial fit parameter:

        * ``poly_fit_a{0-2}``: constants of the polynomial ``p(x) = p[2] * x**deg + p[1]* x + p[0]``


        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data
        is_cft_interval : bool, optional
            ``True`` if the heart rate data passed via ``data`` contains only the CFT Interval,
            ``False`` if it contains the data during the whole CFT procedure. Default: ``False``
        compute_baseline : bool, optional
            ``True`` if Baseline Interval is included in data passed via ``data`` and Baseline heart rate
            should be computed or ``False`` if Baseline heart rate is passed separately via ``hr_baseline``.
            Default: ``True``
        hr_baseline : float, optional
            mean heart rate during Baseline Interval or ``None`` if Baseline interval is present in ``data`` and
            Baseline heart rate is computed from there. Default: ``None``


        Returns
        -------
        dict
            dictionary with CFT polynomial fit parameter

        """
        df_hr_cft, hr_baseline = self._sanitize_cft_input(data, is_cft_interval, compute_baseline, hr_baseline)

        # get time points in seconds
        # TODO check index type
        idx_s = df_hr_cft.index.view(np.int64) / 1e9
        idx_s = idx_s - idx_s[0]

        # apply a 2nd degree polynomial fit
        poly = np.polyfit(idx_s, np.squeeze(df_hr_cft.values), deg=2)
        return {f"poly_fit_a{i}": p for i, p in enumerate(reversed(poly))}

    def _sanitize_cft_input(
        self,
        data: pd.DataFrame,
        is_cft_interval: Optional[bool] = False,
        compute_baseline: Optional[bool] = True,
        hr_baseline: Optional[float] = None,
    ) -> Tuple[pd.DataFrame, float]:
        """Sanitize CFT input.

        Most functions for computing CFT parameter expect multiple possible combinations of input parameter:

        * Either data over the whole duration of the CFT procedure (Baseline, CFT, Recovery):
          Then, the CFT Interval will be extracted and Baseline heart rate will be computed based on the
          ``cft_start`` and ``cft_duration`` parameters of the ``CFT`` object
          (``compute_baseline`` must then be set to ``True`` - the default)
        * Or only data during the CFT interval (``is_cft_interval`` must be set to ``True``).
          Then, the Baseline heart rate muse be explicitly provided via ``hr_baseline`` parameter and
          ``compute_baseline`` must be set to ``False``

        This function sanitizes the input and, independent from the input, always returns a tuple with the data cut
        to the CFT Interval and the mean heart rate during the Baseline Interval.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data
        is_cft_interval : bool, optional
            ``True`` if the heart rate data passed via ``data`` contains only the CFT Interval,
            ``False`` if it contains the data during the whole CFT procedure. Default: ``False``
        compute_baseline : bool, optional
            ``True`` if Baseline Interval is included in data passed via ``data`` and Baseline heart rate
            should be computed or ``False`` if Baseline heart rate is passed separately via ``hr_baseline``.
            Default: ``True``
        hr_baseline : float, optional
            mean heart rate during Baseline Interval or ``None`` if Baseline interval is present in ``data`` and
            Baseline heart rate is computed from there. Default: ``None``

        Returns
        -------
        data_cft : :class:`~pandas.DataFrame`
            heart rate data during CFT Interval

        """
        # extract CFT interval if whole data is passed
        data_cft = data if is_cft_interval else self.extract_cft_interval(data)

        if compute_baseline:
            if is_cft_interval:
                raise ValueError(
                    "`compute_baseline` must be set to False and `baseline_hr` be supplied "
                    "when only CFT data is passed (`is_cft_interval` is `True`)"
                )
            hr_baseline = self.baseline_hr(data)
        else:
            if hr_baseline is None:  # noqa: PLR5501
                raise ValueError("`baseline_hr` must be supplied as parameter when `compute_baseline` is set to False!")

        return data_cft, hr_baseline

    def cft_plot(
        self,
        data: pd.DataFrame,
        **kwargs,
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Draw Cold Face Test (CFT) plot.

        Parameters
        ----------
        data : :class:`~pandas.DataFrame`
            input data
        **kwargs: dict, optional
            optional parameters to be passed to the plot, such as:

            * ``time_baseline`` : duration of Baseline Interval to include in plot or ``None`` to include the
              whole Baseline Interval in the plot.
            * ``time_recovery`` : duration of Recovery Interval to include in plot or ``None`` to include the
              whole Recovery Interval in the plot.
            * ``plot_datetime_index`` : ``True`` to plot x axis with absolute time (:class:`~pandas.DatetimeIndex`),
              or ``False`` to plot data with relative time (starting from second 0). Default: ``False``
            * ``ax``: pre-existing axes for the plot. Otherwise, a new figure and axes object is created and
              returned.
            * ``figsize``: tuple specifying figure dimensions
            * ``ylims``: list to manually specify y axis limits, float to specify y axis margin
              (see :meth:`matplotlib.axes.Axes.margins` for further information), or ``None`` to automatically
              infer y axis limits.
            * ``plot_onset``: whether to plot CFT onset annotations or not: Default: ``True``
            * ``plot_peak_brady``: whether to plot CFT peak bradycardia annotations or not: Default: ``True``
            * ``plot_mean``: whether to plot CFT mean bradycardia annotations or not. Default: ``True``
            * ``plot_baseline``: whether to plot heart rate baseline annotations or not. Default: ``True``
            * ``plot_poly_fit``: whether to plot CFT polynomial fit annotations or not. Default: ``True``

        Returns
        -------
        fig : :class:`~matplotlib.figure.Figure`
            figure object
        ax : :class:`~matplotlib.axes.Axes`
            axes object

        """
        ax: plt.Axes = kwargs.pop("ax", None)

        if ax is None:
            fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
        else:
            fig = ax.get_figure()

        time_baseline = kwargs.get("time_baseline", self.structure.get("Baseline", 0))
        time_recovery = kwargs.get("time_recovery", self.structure.get("Recovery", 0))

        data = data.copy()
        cft_params = self.compute_cft_parameter(data, return_dict=True)

        if not kwargs.get("plot_datetime_index", False):
            data.index = (data.index - data.index[0]).view(np.int64) / 1e9

        times_dict = self._cft_plot_get_cft_times(data, time_baseline, time_recovery)
        df_plot = self._cft_plot_extract_plot_interval(data, times_dict)

        bbox = {
            "fc": (1, 1, 1, plt.rcParams["legend.framealpha"]),
            "ec": plt.rcParams["legend.edgecolor"],
            "boxstyle": "round",
        }

        hr_plot(heart_rate=df_plot, ax=ax, plot_mean=False)
        self._cft_plot_add_phase_annotations(ax, times_dict, **kwargs)
        self._cft_plot_add_param_annotations(data, cft_params, times_dict, ax, bbox, **kwargs)

        self._cft_plot_style_axis(data, ax, **kwargs)

        fig.tight_layout()
        fig.autofmt_xdate(rotation=0, ha="center")
        return fig, ax

    @staticmethod
    def _cft_plot_style_axis(data: pd.DataFrame, ax: plt.Axes, **kwargs):
        ylims = kwargs.get("ylims", None)
        if isinstance(ylims, (tuple, list)):
            ax.set_ylim(ylims)
        else:
            ymargin = 0.2
            if isinstance(ylims, float):
                ymargin = ylims
            ax.margins(x=0, y=ymargin)

        if isinstance(data.index, pd.DatetimeIndex):
            ax.set_xlabel("Time")
        else:
            ax.set_xlabel("Time [s]")

    def _cft_plot_get_cft_times(
        self, data: pd.DataFrame, time_baseline: int, time_recovery: int
    ) -> Dict[str, Union[int, datetime.datetime]]:
        if isinstance(data.index, pd.DatetimeIndex):
            cft_start = data.index[0] + pd.Timedelta(seconds=self.cft_start)
            plot_start = cft_start - pd.Timedelta(seconds=time_baseline)
            cft_end = cft_start + pd.Timedelta(seconds=self.cft_duration)
            plot_end = cft_end + pd.Timedelta(seconds=time_recovery)
        else:
            cft_start = data.index[0] + self.cft_start
            plot_start = cft_start - time_baseline
            cft_end = cft_start + self.cft_duration
            plot_end = cft_end + time_recovery

        cft_times = {
            "plot_start": plot_start,
            "cft_start": cft_start,
            "cft_end": cft_end,
            "plot_end": plot_end,
        }
        return cft_times

    def _cft_plot_extract_plot_interval(
        self, data: pd.DataFrame, times_dict: Dict[str, Union[int, datetime.datetime]]
    ) -> pd.DataFrame:
        plot_start = times_dict["plot_start"]
        plot_end = times_dict["plot_end"]
        if isinstance(data.index, pd.DatetimeIndex):
            df_plot = data.between_time(plot_start.time(), plot_end.time())
        else:
            df_plot = data.loc[plot_start:plot_end]

        return df_plot

    def _cft_plot_add_phase_annotations(
        self, ax: plt.Axes, times_dict: Dict[str, Union[int, datetime.datetime]], **kwargs
    ):
        times = list(zip(list(times_dict.values()), list(times_dict.values())[1:]))
        # filter empty phases, e.g., when no baseline or no recovery phase is present
        times = list(filter(lambda t: (t[-1] - t[0]) != 0, times))

        bg_colors = kwargs.get("background_color", self.cft_plot_params.get("background_color"))
        if bg_colors is None:
            bg_color_base = kwargs.get("background_base_color", self.cft_plot_params.get("background_base_color"))
            bg_colors = list(sns.dark_palette(bg_color_base, n_colors=len(times), reverse=True))
        bg_alphas = kwargs.get("background_alpha", self.cft_plot_params.get("background_alpha"))
        bg_alphas = [bg_alphas] * len(times)
        names = kwargs.get("phase_names", self.cft_plot_params["phase_names"])

        for (start, end), bg_color, bg_alpha, name in zip(times, bg_colors, bg_alphas, names):
            ax.axvspan(xmin=start, xmax=end, color=bg_color, alpha=bg_alpha, lw=0)
            ax.text(
                x=start + 0.5 * (end - start),
                y=0.95,
                transform=ax.get_xaxis_transform(),
                s=name,
                zorder=3,
                ha="center",
                va="center",
            )
        rect = mpatch.Rectangle(
            xy=(0, 0.9),
            width=1,
            height=0.1,
            color="white",
            alpha=0.4,
            zorder=1,
            lw=0,
            transform=ax.transAxes,
        )
        ax.add_patch(rect)

    def _cft_plot_add_param_annotations(self, data, cft_params, times_dict, ax, bbox, **kwargs):
        if kwargs.get("plot_baseline", True):
            self._cft_plot_add_baseline(cft_params, times_dict, ax)
        if kwargs.get("plot_mean", True):
            self._cft_plot_add_mean_bradycardia(data, cft_params, times_dict, ax, bbox)
        if kwargs.get("plot_onset", True):
            self._cft_plot_add_onset(data, cft_params, times_dict, ax, bbox)
        if kwargs.get("plot_peak_brady", True):
            self._cft_plot_add_peak_bradycardia(data, cft_params, times_dict, ax, bbox)
        if kwargs.get("plot_poly_fit", True):
            self._cft_plot_add_poly_fit(data, cft_params, ax)

    def _cft_plot_add_peak_bradycardia(
        self,
        data: pd.DataFrame,
        cft_params: Dict,
        cft_times: Dict,
        ax: plt.Axes,
        bbox: Dict,
    ) -> None:

        color_key = "fau"
        if isinstance(data.index, pd.DatetimeIndex):
            brady_loc = cft_params["peak_brady"]
            brady_x = brady_loc
            brady_y = float(data.loc[brady_loc])
        else:
            brady_loc = cft_params["cft_start_idx"] + cft_params["peak_brady_idx"]
            brady_x = data.index[brady_loc]
            brady_y = float(data.iloc[brady_loc])

        hr_baseline = cft_params["baseline_hr"]
        max_hr_cft = float(self.extract_cft_interval(data).max())
        cft_start = cft_times["cft_start"]

        color = getattr(colors_all, color_key)
        color_adjust = getattr(colors_all, f"{color_key}_dark")

        # Peak Bradycardia vline
        ax.axvline(x=brady_x, ls="--", lw=2, alpha=0.6, color=color)

        # Peak Bradycardia marker
        ax.plot(
            brady_x,
            brady_y,
            color=color,
            marker="o",
            markersize=7,
        )

        # Peak Bradycardia hline
        xmax = brady_x + pd.Timedelta(seconds=20) if isinstance(data.index, pd.DatetimeIndex) else brady_x + 20

        ax.hlines(
            y=brady_y,
            xmin=brady_x,
            xmax=xmax,
            ls="--",
            lw=2,
            color=color_adjust,
            alpha=0.6,
        )

        # Peak Bradycardia arrow
        if isinstance(data.index, pd.DatetimeIndex):
            brady_x_offset = brady_x + pd.Timedelta(seconds=10)
        else:
            brady_x_offset = brady_x + 10

        ax.annotate(
            "",
            xy=(brady_x_offset, brady_y),
            xytext=(brady_x_offset, hr_baseline),
            arrowprops={
                "arrowstyle": "<->",
                "lw": 2,
                "color": color_adjust,
                "shrinkA": 0.0,
                "shrinkB": 0.0,
            },
        )

        # Peak Bradycardia Text
        ax.annotate(
            "$Peak_{CFT}$: " + "{:.1f} %".format(cft_params["peak_brady_percent"]),
            xy=(brady_x_offset, brady_y),
            xytext=(10, -5),
            textcoords="offset points",
            bbox=bbox,
            ha="left",
            va="top",
        )

        # Peak Bradycardia Latency arrow
        ax.annotate(
            "",
            xy=(cft_start, max_hr_cft),
            xytext=(brady_x, max_hr_cft),
            arrowprops={
                "arrowstyle": "<->",
                "lw": 2,
                "color": color,
                "shrinkA": 0.0,
                "shrinkB": 0.0,
            },
        )

        # Peak Bradycardia Latency Text
        ax.annotate(
            "$Latency_{CFT}$: " + "{:.1f} s".format(cft_params["peak_brady_latency"]),
            xy=(brady_x, max_hr_cft),
            xytext=(-7.5, 10),
            textcoords="offset points",
            bbox=bbox,
            ha="right",
            va="bottom",
        )

    def _cft_plot_add_baseline(
        self,
        cft_params: Dict,
        cft_times: Dict,
        ax: plt.Axes,
    ) -> None:
        color_key = "tech"

        # Baseline HR
        ax.hlines(
            y=cft_params["baseline_hr"],
            xmin=cft_times["plot_start"],
            xmax=cft_times["cft_end"],
            ls="--",
            lw=2,
            color=getattr(colors_all, color_key),
            alpha=0.6,
        )

    def _cft_plot_add_mean_bradycardia(
        self,
        data: pd.DataFrame,
        cft_params: Dict,
        cft_times: Dict,
        ax: plt.Axes,
        bbox: Dict,
    ) -> None:

        color_key = "wiso"
        mean_hr = cft_params["mean_hr_bpm"]
        cft_start = cft_times["cft_start"]
        cft_end = cft_times["cft_end"]

        # Mean HR during CFT
        ax.hlines(
            y=mean_hr,
            xmin=cft_start,
            xmax=cft_end,
            ls="--",
            lw=2,
            color=getattr(colors_all, color_key),
            alpha=0.6,
        )

        x_offset = cft_end - pd.Timedelta(seconds=5) if isinstance(data.index, pd.DatetimeIndex) else cft_end - 5

        # Mean Bradycardia arrow
        ax.annotate(
            "",
            xy=(x_offset, mean_hr),
            xytext=(x_offset, cft_params["baseline_hr"]),
            arrowprops={
                "arrowstyle": "<->",
                "lw": 2,
                "color": getattr(colors_all, f"{color_key}_dark"),
                "shrinkA": 0.0,
                "shrinkB": 0.0,
            },
        )

        # Mean Bradycardia Text
        ax.annotate(
            "$Mean_{CFT}$: " + "{:.1f} %".format(cft_params["mean_brady_percent"]),
            xy=(x_offset, mean_hr),
            xytext=(-5, -5),
            textcoords="offset points",
            bbox=bbox,
            ha="right",
            va="top",
        )

    def _cft_plot_add_onset(
        self,
        data: pd.DataFrame,
        cft_params: Dict,
        cft_times: Dict,
        ax: plt.Axes,
        bbox: Dict,
    ) -> None:

        color_key = "med"
        color = getattr(colors_all, color_key)

        if isinstance(data.index, pd.DatetimeIndex):
            onset_idx = cft_params["onset"]
            onset_x = onset_idx
            onset_y = float(data.loc[onset_idx])
        else:
            onset_idx = cft_params["cft_start_idx"] + cft_params["onset_idx"]
            onset_y = float(data.iloc[onset_idx])
            onset_x = data.index[onset_idx]

        # CFT Onset vline
        ax.axvline(onset_x, ls="--", lw=2, alpha=0.6, color=color)

        # CFT Onset marker
        ax.plot(
            onset_x,
            onset_y,
            color=color,
            marker="o",
            markersize=7,
        )

        # CFT Onset arrow
        ax.annotate(
            "",
            xy=(onset_x, onset_y),
            xytext=(cft_times["cft_start"], onset_y),
            arrowprops={
                "arrowstyle": "<->",
                "lw": 2,
                "color": color,
                "shrinkA": 0.0,
                "shrinkB": 0.0,
            },
        )

        # CFT Onset Text
        ax.annotate(
            "$Onset_{CFT}$: " + "{:.1f} s".format(cft_params["onset_latency"]),
            xy=(onset_x, onset_y),
            xytext=(-10, -10),
            textcoords="offset points",
            bbox=bbox,
            ha="right",
            va="top",
        )

    def _cft_plot_add_poly_fit(
        self,
        data: pd.DataFrame,
        cft_params: Dict,
        ax: plt.Axes,
    ) -> None:

        color_key = "phil"
        df_cft = self.extract_cft_interval(data)
        if isinstance(df_cft.index, pd.DatetimeIndex):
            x_poly = df_cft.index.view(np.int64) / 1e9
        else:
            x_poly = df_cft.index

        x_poly = x_poly - x_poly[0]
        y_poly = (
            cft_params["poly_fit_a2"] * x_poly**2 + cft_params["poly_fit_a1"] * x_poly + cft_params["poly_fit_a0"]
        )

        ax.plot(
            df_cft.index,
            y_poly,
            lw=2,
            color=getattr(colors_all, color_key),
            alpha=0.6,
            zorder=2,
        )



"""Module providing input/output functions."""
from biopsykit.io import carwatch_logs, ecg, eeg, nilspod, saliva, sleep, sleep_analyzer
from biopsykit.io.io import (
    convert_time_log_datetime,
    convert_time_log_dict,
    load_atimelogger_file,
    load_codebook,
    load_long_format_csv,
    load_pandas_dict_excel,
    load_questionnaire_data,
    load_subject_condition_list,
    load_time_log,
    write_pandas_dict_excel,
    write_result_dict,
)

__all__ = [
    "load_long_format_csv",
    "load_time_log",
    "load_atimelogger_file",
    "load_subject_condition_list",
    "load_questionnaire_data",
    "load_pandas_dict_excel",
    "load_codebook",
    "convert_time_log_datetime",
    "convert_time_log_dict",
    "write_pandas_dict_excel",
    "write_result_dict",
    "carwatch_logs",
    "ecg",
    "eeg",
    "nilspod",
    "saliva",
    "sleep",
    "sleep_analyzer",
]



"""Module for computing Rest Periods from raw acceleration signals."""
import datetime
from typing import Union

import numpy as np
import pandas as pd
from biopsykit.utils._datatype_validation_helper import _assert_num_columns
from biopsykit.utils._types import arr_t
from biopsykit.utils.array_handling import sliding_window


class RestPeriods:
    """Compute Rest Periods from raw acceleration signals.

    Rest periods are periods with large inactivity, characterized by low angle changes in the acceleration signal.

    The longest rest period, the *Major Rest Period*, is used to determine the sleep window (i.e., time in bed)
    of the day.


    References
    ----------
    van Hees, V. T., Sabia, S., Anderson, K. N., Denton, S. J., Oliver, J., Catt, M., Abell, J. G., Kivimäki, M.,
    Trenell, M. I., & Singh-Manoux, A. (2015). A Novel, Open Access Method to Assess Sleep Duration Using a
    Wrist-Worn Accelerometer. *PLoS ONE*, 10(11), 1-13. https://doi.org/10.1371/journal.pone.0142533

    """

    sampling_rate: float

    def __init__(self, sampling_rate: float):
        """Initialize a new ``RestPeriods`` instance.

        Parameters
        ----------
        sampling_rate : float
            sampling rate of recorded data in Hz

        """
        self.sampling_rate = sampling_rate

    def predict(self, data: arr_t) -> pd.DataFrame:
        """Predict Rest Periods from acceleration data.

        Parameters
        ----------
        data : array_like
            input acceleration data. Must be 3-d.

        Returns
        -------
        :class:`~pandas.DataFrame`
            dataframe with start, end, and total duration of each rest period detected by the algorithm

        """
        data = data.filter(like="acc") if isinstance(data, pd.DataFrame) else pd.DataFrame(data)

        _assert_num_columns(data, 3)

        # rolling median 5 second
        data = data.rolling(int(5 * self.sampling_rate), min_periods=0).median()

        # get angle
        angle = np.arctan(data["acc_z"] / ((data["acc_x"] ** 2 + data["acc_y"] ** 2) ** 0.5)) * (180.0 / np.pi)

        window_s = 5  # 5 seconds
        overlap = 0

        angle_sliding = sliding_window(
            angle, window_sec=window_s, sampling_rate=self.sampling_rate, overlap_percent=overlap
        )

        index_resample = self._resample_index(data, window_s, overlap)

        df_angle = pd.DataFrame(
            np.abs(np.diff(np.nanmean(angle_sliding, axis=1))), columns=["angle"], index=index_resample[1:]
        )
        df_angle = df_angle.rolling(60).median()  # rolling median, 60 * 5 seconds per sample = 5 minutes

        minimum_rest_threshold = 0.0
        maximum_rest_threshold = 1000.0

        # calculate and apply threshold
        thresh = np.min(
            [
                np.max([np.percentile(df_angle["angle"].dropna().values, 10) * 15.0, minimum_rest_threshold]),
                maximum_rest_threshold,
            ]
        )

        df_angle[df_angle < thresh] = 0.0
        df_angle[df_angle >= thresh] = 1.0

        minimum_rest_block = 30
        allowed_rest_break = 60

        # drop rest blocks < minimum_rest_block minutes (except first and last)
        df_angle["block"] = (df_angle["angle"].diff().ne(0)).cumsum()
        groups = list(df_angle.groupby(by="block"))
        # exclude first and last rest block
        for _, group in groups[1:-1]:
            if group["angle"].sum() == 0 and len(group) < (12 * minimum_rest_block):
                # 5 second intervals => 12x for 1min
                df_angle.loc[group.index[0] : group.index[-1], "angle"] = 1

        # drop active blocks < allowed_rest_break minutes (except first and last)
        df_angle["block"] = (df_angle["angle"].diff().ne(0)).cumsum()
        groups = list(df_angle.groupby(by="block"))
        for _, group in groups[1:-1]:
            if group["angle"].sum() == len(group) and len(group) < (12 * allowed_rest_break):
                # 5 second intervals => 12x for 1min
                df_angle.loc[group.index[0] : group.index[-1], "angle"] = 0

        # get longest block
        df_angle["block"] = (df_angle["angle"].diff().ne(0)).cumsum()
        group = df_angle[df_angle.angle == 0].groupby("block")
        grp_max = group.get_group(group.size().idxmax())

        total_duration = data.index[-1] - data.index[0]

        return self._major_rest_period(data, total_duration, grp_max)

    def _resample_index(self, data: pd.DataFrame, window_s: int, overlap: float):
        index_resample = sliding_window(
            data.index.values, window_sec=window_s, sampling_rate=self.sampling_rate, overlap_percent=overlap
        )[:, 0]

        if isinstance(data.index, pd.DatetimeIndex):
            index_resample = pd.DatetimeIndex(index_resample)
            index_resample = index_resample.tz_localize("UTC").tz_convert(data.index.tzinfo)
        return index_resample

    def _major_rest_period(
        self, data: pd.DataFrame, total_duration: Union[datetime.timedelta, float], grp_max: pd.DataFrame
    ) -> pd.DataFrame:
        if isinstance(data.index, pd.DatetimeIndex):
            total_duration = total_duration.total_seconds() / 3600.0
            start = grp_max.index[0]
            end = grp_max.index[-1]
        else:
            total_duration = (total_duration / self.sampling_rate) / 3600.0
            start = int(grp_max.index[0] / (self.sampling_rate * 60))
            end = int(grp_max.index[-1] / (self.sampling_rate * 60))

        mrp = {"start": start, "end": end, "total_duration": total_duration}
        return pd.DataFrame(mrp, index=[0])



"""Module with functions to process IMU data."""
from typing import Optional, Union

import numpy as np
import pandas as pd
from biopsykit.utils._types import arr_t
from biopsykit.utils.array_handling import sliding_window
from biopsykit.utils.datatype_helper import Acc3dDataFrame, Gyr3dDataFrame, ImuDataFrame
from biopsykit.utils.time import utc


def convert_acc_data_to_g(
    data: Union[Acc3dDataFrame, ImuDataFrame], inplace: Optional[bool] = False
) -> Optional[Union[Acc3dDataFrame, ImuDataFrame]]:
    """Convert acceleration data from :math:`m/s^2` to g.

    Parameters
    ----------
    data : :class:`~biopsykit.utils.datatype_helper.AccDataFrame` or \
            :class:`~biopsykit.utils.datatype_helper.ImuDataFrame`
        dataframe containing acceleration data.
    inplace : bool, optional
        whether to perform the operation inplace or not. Default: ``False``

    Returns
    -------
    :class:`~biopsykit.utils.datatype_helper.AccDataFrame` or :class:`~biopsykit.utils.datatype_helper.ImuDataFrame`
        acceleration data converted to g

    """
    if not inplace:
        data = data.copy()
    acc_cols = data.filter(like="acc").columns
    data.loc[:, acc_cols] = data.loc[:, acc_cols] / 9.81

    if inplace:
        return None
    return data


def sliding_windows_imu(
    data: arr_t,
    window_samples: Optional[int] = None,
    window_sec: Optional[int] = None,
    sampling_rate: Optional[float] = 0,
    overlap_samples: Optional[int] = None,
    overlap_percent: Optional[float] = None,
) -> pd.DataFrame:
    """Create sliding windows from IMU data.

    The window size of sliding windows can either be specified in *samples* (``window_samples``)
    or in *seconds* (``window_sec``, together with ``sampling_rate``).

    The overlap of windows can either be specified in *samples* (``overlap_samples``)
    or in *percent* (``overlap_percent``).

    If the input data has the shape (m, n) the output data will have the following shape:
    (m * window_size, num_windows), where x depends on the number of data sources (acc, gyr, or both) and the
    axes (x, y, z). The number of windows depends on the input length of the array (n), the window size, and the
    window overlap.

    The output data will be a :class:`~pandas.DataFrame` where the windowed values of each axes are concatenated
    to one row. This means that the output data has a multi-level column index with two levels if the input data only
    has one column level (``axis``):  ``axis`` ("x", "y", "z") as the first level and ``samples`` (array index within
    the window) as second level. If the input data has two column levels (``channel`` and ``axis``) the output
    dataframe will have three levels: ``channel``, ``axis``, and ``samples``.


    Parameters
    ----------
    data : array_like
        input data
    window_samples : int, optional
        window size in samples or ``None`` if window size is specified in seconds + sampling rate. Default: ``None``
    window_sec : int, optional
        window size in seconds or ``None`` if window size is specified in samples. Default: ``None``
    sampling_rate : float, optional
        sampling rate of data in Hz. Only needed if window size is specified in seconds (``window_sec`` parameter).
        Default: ``None``
    overlap_samples : int, optional
        overlap of windows in samples or ``None`` if window overlap is specified in percent. Default: ``None``
    overlap_percent : float, optional
        overlap of windows in percent or ``None`` if window overlap is specified in samples. Default: ``None``


    Returns
    -------
    :class:`~pandas.DataFrame`
        sliding windows from input data


    See Also
    --------
    :func:`~biopsykit.utils.array_handling.sliding_window_view`
        create sliding window of input array. low-level function with less input parameter configuration possibilities

    """
    index = None
    index_resample = None
    if isinstance(data, (pd.DataFrame, pd.Series)):
        index = data.index

    data_window = sliding_window(
        data,
        window_samples=window_samples,
        window_sec=window_sec,
        sampling_rate=sampling_rate,
        overlap_samples=overlap_samples,
        overlap_percent=overlap_percent,
    )
    if index is not None:
        index_resample = sliding_window(
            index.values,
            window_samples=window_samples,
            window_sec=window_sec,
            sampling_rate=sampling_rate,
            overlap_samples=overlap_samples,
            overlap_percent=overlap_percent,
        )[:, 0]
        if isinstance(index, pd.DatetimeIndex):
            index_resample = pd.DatetimeIndex(index_resample)
            index_resample = index_resample.tz_localize(utc).tz_convert(index.tzinfo)
        else:
            index_resample = pd.Index(index_resample)
        index_resample = index_resample.set_names(index.names)

    data_window = np.transpose(data_window)
    data_window = {
        axis: pd.DataFrame(np.transpose(data), index=index_resample) for axis, data in zip(data.columns, data_window)
    }
    data_window = pd.concat(data_window, axis=1)
    if data_window.columns.nlevels == 2:
        data_window.columns.names = ["axis", "samples"]
    elif data_window.columns.nlevels == 3:
        data_window.columns.names = ["channel", "axis", "samples"]
    return data_window


def var_norm_windows(data: Union[Acc3dDataFrame, Gyr3dDataFrame]) -> pd.DataFrame:
    r"""Compute the norm of the variance of each axis for a windowed signal.

    This function computes the norm of variance according to:

    .. math::
        var_{norm} = \sqrt{var_x^2 + var_y^2 + var_z^2}

    where :math:`var_i` is the variance of axis :math:`i` in the window

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        input data, split into windows using :func:`~biopsykit.signals.imu.sliding_windows_imu`

    Returns
    -------
    :class:`~pandas.DataFrame`
        norm of variance for windowed signal

    """
    var = data.groupby(level="axis", axis=1).apply(lambda x: np.var(x, axis=1))
    norm = pd.DataFrame(np.linalg.norm(var, axis=1), index=var.index, columns=["var_norm"])
    return norm



"""Module representing the Cortisol Awakening Response (CAR) protocol."""
from typing import Optional, Tuple

import matplotlib.pyplot as plt
from biopsykit.plotting import lineplot
from biopsykit.protocols import BaseProtocol
from biopsykit.utils.datatype_helper import is_saliva_raw_dataframe


class CAR(BaseProtocol):
    """Class representing psychological protocols for assessing the cortisol awakening response (CAR)."""

    def __init__(self, name: Optional[str] = None, **kwargs):
        """Class representing psychological protocols for assessing the cortisol awakening response (CAR).

        Parameters
        ----------
        name : str, optional
            name of CAR study or ``None`` to use default name ("CAR"). Default: ``None``
        **kwargs
            additional parameters to be passed to ``CAR`` and its superclass, ``BaseProtocol``, such as:

            * ``car_saliva_plot_params``: parameters to style :meth:`~biopsykit.protocols.CAR.car_saliva_plot`

        """
        if name is None:
            name = "CAR"

        car_saliva_plot_params = {"xlabel": "Time after Awakening [min]", "ylabel": r"Cortisol [nmol/l]"}
        car_saliva_plot_params.update(kwargs.pop("car_saliva_plot_params", {}))
        self.car_plot_params = car_saliva_plot_params
        """Plot parameters to style :meth:`~biopsykit.protocols.CAR.car_saliva_plot`."""

        super().__init__(name, **kwargs)

    def car_saliva_plot(self, saliva_type: Optional[str] = "cortisol", **kwargs) -> Tuple[plt.Figure, plt.Axes]:
        """Plot CAR saliva data as lineplot.

        Parameters
        ----------
        saliva_type : str, optional
            type of saliva data to plot. Default: ``cortisol``
        **kwargs : optional arguments to be passed to :func:`~biopsykit.plotting.lineplot`


        Returns
        -------
        fig : :class:`~matplotlib.figure.Figure`
            figure object
        ax : :class:`~matplotlib.axes.Axes`
            axes object


        See Also
        --------
        :func:`~biopsykit.plotting.lineplot`
            draw line plot with error bars

        """
        data = self.saliva_data[saliva_type]
        is_saliva_raw_dataframe(data, saliva_type)

        kwargs.setdefault("x", "sample")
        kwargs.setdefault("y", saliva_type)
        kwargs.setdefault("xticklabels", kwargs.get("sample_times", self.sample_times[saliva_type]))
        if len(data.index.get_level_values(kwargs.get("x")).unique()) != len(kwargs.get("xticklabels")):
            raise ValueError(
                "If samples have individual sample times for each subject, sample times for plotting must "
                "explicitly be provided by the 'sample_times' argument!"
            )
        kwargs.update(self.car_plot_params)
        return lineplot(data, **kwargs)



"""Module providing functions for plotting ECG data."""
from typing import TYPE_CHECKING, Dict, List, Optional, Sequence, Tuple

import matplotlib as mpl
import matplotlib.dates as mdates
import matplotlib.gridspec as gs
import matplotlib.pyplot as plt
import matplotlib.ticker as mticks
import neurokit2 as nk
import numpy as np
import pandas as pd
import seaborn as sns
from biopsykit.signals.ecg.ecg import _assert_ecg_input
from biopsykit.utils.array_handling import sanitize_input_1d
from biopsykit.utils.datatype_helper import EcgResultDataFrame
from fau_colors import cmaps, colors_all
from neurokit2.hrv.hrv_frequency import _hrv_frequency_show
from neurokit2.hrv.hrv_utils import _hrv_get_rri

if TYPE_CHECKING:
    from biopsykit.signals.ecg import EcgProcessor

__all__ = [
    "ecg_plot",
    "hr_plot",
    "hrv_plot",
    "hr_distribution_plot",
    "rr_distribution_plot",
    "hrv_frequency_plot",
    "hrv_poincare_plot",
    "individual_beats_plot",
]

# TODO add signal plot method for all phases


def ecg_plot(
    ecg_processor: Optional["EcgProcessor"] = None,
    key: Optional[str] = None,
    ecg_signal: Optional[pd.DataFrame] = None,
    heart_rate: Optional[pd.DataFrame] = None,
    sampling_rate: Optional[int] = 256,
    plot_ecg_signal: Optional[bool] = True,
    plot_distribution: Optional[bool] = True,
    plot_individual_beats: Optional[bool] = True,
    **kwargs,
) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
    """Plot ECG processing results.

    By default, this plot consists of four subplots:

    * `top left`: course of ECG signal plot with signal quality indicator,
      detected R peaks and R peaks marked as outlier
    * `bottom left`: course of heart rate (tachogram)
    * `top right`: individual heart beats overlaid on top of each other
    * `bottom right`: heart rate distribution (histogram)


    To use this function, either simply pass an :class:`~biopsykit.signals.ecg.EcgProcessor` object together with
    a ``key`` indicating which phase needs to be processed should be processed or the two dataframes ``ecg_signal``
    and ``heart_rate`` resulting from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.


    Parameters
    ----------
    ecg_processor : :class:`~biopsykit.signals.ecg.EcgProcessor`, optional
        ``EcgProcessor`` object. If this argument is supplied, the ``key`` argument needs to be supplied as well.
    key : str, optional
        Dictionary key of the phase to process. Needed when ``ecg_processor`` is passed as argument.
    ecg_signal : :class:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`, optional
        Dataframe with processed ECG signal. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
    heart_rate : :class:`~pandas.DataFrame`, optional
        Dataframe with heart rate output. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
    sampling_rate : float, optional
        Sampling rate of recorded data in Hz. Not needed if ``ecg_processor`` is supplied as parameter.
        Default: 256
    plot_ecg_signal : bool, optional
        Whether to plot the cleaned ECG signal in a subplot or not. Default: ``True``
    plot_distribution : bool, optional
        Whether to plot the heart rate distribution in a subplot or not. Default: ``True``
    plot_individual_beats : bool, optional
        Whether to plot the individual heart beats in a subplot or not. Default: ``True``
    **kwargs
        Additional parameters to configure the plot. Parameters include:

        * ``figsize``: Figure size
        * ``title``: Optional name to add to plot title (after "Electrocardiogram (ECG)")
        * ``legend_loc``: Location of legend in plot. Passed as `loc` parameter to :meth:`matplotlib.axes.Axes.legend`.
        * ``legend_fontsize``: Fontsize of legend text. Passed as `fontsize` parameter to
          :meth:`matplotlib.axes.Axes.legend`.


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        Figure object
    axs : list of :class:`~matplotlib.axes.Axes`
        list of subplot axes objects


    See Also
    --------
    :func:`~biopsykit.signals.ecg.plotting.hr_plot`
        plot heart rate only
    :func:`~biopsykit.signals.ecg.plotting.hr_distribution_plot`
        plot heart rate distribution only
    :func:`~biopsykit.signals.ecg.plotting.individual_beats_plot`
        plot individual beats only
    :func:`~biopsykit.signals.ecg.plotting.hrv_plot`
        plot heart rate variability

    """
    _assert_ecg_input(ecg_processor, key, ecg_signal, heart_rate)
    if ecg_processor is not None:
        ecg_signal = ecg_processor.ecg_result[key]
        heart_rate = ecg_processor.heart_rate[key]
        sampling_rate = ecg_processor.sampling_rate

    sns.set_palette(cmaps.faculties)

    title = kwargs.get("title", None)
    _set_plt_rcparams(ecg_signal)

    # Prepare figure and set axes
    fig = plt.figure(figsize=kwargs.get("figsize"), constrained_layout=False)
    axs = _get_ecg_plot_specs(fig, plot_ecg_signal, plot_individual_beats, plot_distribution)
    _ecg_plot_set_title(fig, title)

    plt.subplots_adjust(hspace=0.3, wspace=0.1)

    peaks = np.where(ecg_signal["ECG_R_Peaks"] == 1)[0]
    outlier = np.array([])
    if "R_Peak_Outlier" in ecg_signal:
        outlier = np.where(ecg_signal["R_Peak_Outlier"] == 1)[0]

    peaks = np.setdiff1d(peaks, outlier)

    if plot_ecg_signal:
        _ecg_plot(axs, ecg_signal, peaks, outlier, **kwargs)

    # Plot heart rate: plot outlier only if no ECG signal is plotted
    hr_plot(heart_rate, plot_outlier=not plot_ecg_signal, outlier=ecg_signal.index[outlier], ax=axs["hr"])
    axs["hr"].set_xlabel("Time")

    # Plot individual heart beats
    if plot_individual_beats:
        individual_beats_plot(ecg_signal, peaks, sampling_rate, ax=axs["beats"])

    # Plot heart rate distribution
    if plot_distribution:
        hr_distribution_plot(heart_rate, ax=axs["dist"])

    fig.tight_layout()
    return fig, list(axs.values())


def _set_plt_rcparams(ecg_signal: pd.DataFrame):
    if isinstance(ecg_signal.index, pd.DatetimeIndex):
        plt.rcParams["timezone"] = ecg_signal.index.tz.zone
    plt.rcParams["mathtext.default"] = "regular"


def _ecg_plot_set_title(fig: plt.Figure, title: str):
    if title:
        fig.suptitle(f"Electrocardiogram (ECG) - {title}", fontweight="bold")
    else:
        fig.suptitle("Electrocardiogram (ECG)", fontweight="bold")


def _get_ecg_plot_specs(  # pylint:disable=too-many-branches
    fig: plt.Figure, plot_ecg_signal: bool, plot_individual_beats: bool, plot_distribution: bool
) -> Dict[str, plt.Axes]:
    if plot_individual_beats or plot_distribution:
        spec = gs.GridSpec(2, 2, width_ratios=[3, 1])
        if plot_ecg_signal:
            axs = {
                "ecg": fig.add_subplot(spec[0, :-1]),
                "hr": fig.add_subplot(spec[1, :-1]),
            }
        else:
            axs = {"hr": fig.add_subplot(spec[:, :-1])}
        if plot_distribution and plot_individual_beats:
            axs["beats"] = fig.add_subplot(spec[0, -1])
            axs["dist"] = fig.add_subplot(spec[1, -1])
        elif plot_individual_beats:
            axs["beats"] = fig.add_subplot(spec[:, -1])
        elif plot_distribution:
            axs["dist"] = fig.add_subplot(spec[:, -1])
    elif plot_ecg_signal:
        axs = {"ecg": fig.add_subplot(2, 1, 1), "hr": fig.add_subplot(2, 1, 2)}
    else:
        axs = {"hr": fig.add_subplot(1, 1, 1)}
    return axs


def _ecg_plot(
    axs: Dict[str, plt.Axes],
    ecg_signal: EcgResultDataFrame,
    peaks: np.array,
    outlier: np.array,
    **kwargs,
):
    legend_loc = kwargs.get("legend_loc", "upper right")
    legend_fontsize = kwargs.get("legend_fontsize", "small")

    axs["ecg"].get_shared_x_axes().join(axs["ecg"], axs["hr"])

    # z-normalize the ecg signal for better visualization
    ecg_clean = nk.standardize(ecg_signal["ECG_Clean"])
    x_axis = ecg_signal.index
    ylim_ecg = [-5, 10]
    quality = ecg_signal["ECG_Quality"] * ylim_ecg[1]
    minimum_line = np.full(len(x_axis), ylim_ecg[0])

    # Plot quality area first
    axs["ecg"].fill_between(
        x_axis,
        minimum_line,
        quality,
        alpha=0.2,
        zorder=2,
        interpolate=True,
        facecolor=colors_all.med,
        label="Quality",
    )
    # Plot signals
    axs["ecg"].plot(
        ecg_clean,
        color=colors_all.fau,
        label="ECG (z-norm.)",
        zorder=1,
        linewidth=1.5,
    )
    axs["ecg"].scatter(
        x_axis[peaks],
        ecg_clean.iloc[peaks],
        color=colors_all.nat,
        label="R Peaks",
        zorder=2,
    )
    if "R_Peak_Outlier" in ecg_signal:
        axs["ecg"].scatter(
            x_axis[outlier],
            ecg_clean[outlier],
            color=colors_all.phil,
            label="Outlier",
            zorder=2,
        )
    axs["ecg"].set_ylabel("ECG Signal (z-norm.)")
    axs["ecg"].set_ylim(ylim_ecg)

    # Optimize legend
    handles, labels = axs["ecg"].get_legend_handles_labels()
    # order = [2, 0, 1, 3]
    order = [0, 1, 2, 3] if "R_Peak_Outlier" in ecg_signal else [0, 1, 2]

    axs["ecg"].legend(
        [handles[idx] for idx in order], [labels[idx] for idx in order], loc=legend_loc, fontsize=legend_fontsize
    )

    axs["ecg"].tick_params(axis="x", which="both", bottom=True, labelbottom=True)
    axs["ecg"].tick_params(axis="y", which="major", left=True)

    if isinstance(ecg_signal.index, pd.DatetimeIndex):
        # TODO add axis style for non-Datetime axes
        axs["ecg"].xaxis.set_major_locator(mdates.MinuteLocator())
        axs["ecg"].xaxis.set_major_formatter(mdates.DateFormatter("%H:%M"))
        axs["ecg"].xaxis.set_minor_locator(mticks.AutoMinorLocator(6))


def hr_plot(
    heart_rate: pd.DataFrame,
    plot_mean: Optional[bool] = True,
    plot_outlier: Optional[bool] = False,
    outlier: Optional[np.ndarray] = None,
    **kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot course of heart rate over time (tachogram).

    This plot is also used as subplot in :func:`~biopsykit.signals.ecg.plotting.ecg_plot`.


    Parameters
    ----------
    heart_rate : :class:`~pandas.DataFrame`
        Dataframe with heart rate output. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
    plot_mean : bool, optional
        Whether to plot the mean heart rate as horizontal line or not. Default: ``True``
    plot_outlier : bool, optional
        Whether to plot ECG signal outlier as vertical outlier or not. Default: ``False``
    outlier : :class:`~numpy.ndarray`, optional
        List of outlier indices. Only needed if ``plot_outlier`` is ``True``. Default: ``None``
    **kwargs
        Additional parameters to configure the plot. Parameters include:

        * ``figsize``: Figure size
        * ``title``: Optional name to add to plot title (after "Electrocardiogram (ECG)")
        * ``legend_loc``: Location of legend in plot. Passed as `loc` parameter to :meth:`matplotlib.axes.Axes.legend`.
        * ``legend_fontsize``: Fontsize of legend text. Passed as `fontsize` parameter to
          :meth:`matplotlib.axes.Axes.legend`.
        * ``ax``: Pre-existing axes for the plot. Otherwise, a new figure and axes object are created and returned.


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object


    See Also
    --------
    :func:`~biopsykit.signals.ecg.plotting.ecg_plot`
        plot ECG overview

    """
    ax: plt.Axes = kwargs.pop("ax", None)
    title: str = kwargs.get("title", None)
    legend_loc = kwargs.get("legend_loc", "upper right")
    legend_fontsize = kwargs.get("legend_fontsize", "small")
    plt.rcParams["mathtext.default"] = "regular"

    color = kwargs.pop("color", colors_all.wiso)
    mean_color = kwargs.pop("mean_color", colors_all.wiso_dark)

    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    else:
        fig = ax.get_figure()

    sns.set_palette(cmaps.faculties)

    if title:
        ax.set_title(f"Heart Rate - {title}")

    ax.plot(heart_rate["Heart_Rate"], color=color, label="Heart Rate", linewidth=1.5, **kwargs)

    if plot_mean:
        _hr_plot_plot_mean(heart_rate, mean_color, ax)

    ax.set_ylim(auto=True)

    if plot_outlier:
        _hr_plot_plot_outlier(heart_rate, outlier, ax)

    _hr_plot_style_axis(heart_rate, ax)

    if plot_mean or plot_outlier:
        ax.legend(loc=legend_loc, fontsize=legend_fontsize)

    fig.tight_layout()
    fig.autofmt_xdate(rotation=0, ha="center")
    return fig, ax


def _hr_plot_plot_mean(heart_rate: pd.DataFrame, mean_color: str, ax: plt.Axes):
    rate_mean = heart_rate["Heart_Rate"].mean()
    ax.axhline(
        y=rate_mean,
        label=f"Mean: {rate_mean:.1f} bpm",
        linestyle="--",
        color=mean_color,
        linewidth=2,
    )
    ax.margins(x=0)


def _hr_plot_plot_outlier(heart_rate: pd.DataFrame, outlier: np.ndarray, ax: plt.Axes):
    if "R_Peak_Outlier" in heart_rate.columns:
        outlier = heart_rate["R_Peak_Outlier"]
        outlier = heart_rate.index[np.where(outlier == 1)[0]]
    if outlier is not None:
        ax.scatter(
            x=outlier,
            y=heart_rate.loc[outlier, "Heart_Rate"],
            color=colors_all.phil,
            zorder=3,
            label="ECG Outlier",
        )
        ax.relim()


def _hr_plot_style_axis(heart_rate: pd.DataFrame, ax: plt.Axes):
    if isinstance(heart_rate.index, pd.DatetimeIndex):
        # TODO add axis style for non-Datetime axes
        ax.xaxis.set_major_locator(mdates.MinuteLocator())
        ax.xaxis.set_major_formatter(mdates.DateFormatter("%H:%M"))
        ax.xaxis.set_minor_locator(mticks.AutoMinorLocator(6))

    ax.tick_params(axis="x", which="both", bottom=True)
    ax.tick_params(axis="y", which="major", left=True)
    ax.yaxis.set_major_locator(mticks.MaxNLocator(5, steps=[5, 10]))
    ax.set_xlabel("Time")
    ax.set_ylabel("Heart Rate [bpm]")


def hrv_plot(
    ecg_processor: Optional["EcgProcessor"] = None,
    key: Optional[str] = None,
    ecg_signal: Optional[pd.DataFrame] = None,
    rpeaks: Optional[pd.DataFrame] = None,
    sampling_rate: Optional[int] = 256,
    plot_psd: Optional[bool] = True,
    **kwargs,
) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
    """Plot Heart Rate Variability results.

    By default, it consists of 3 plots:

    * `top left`: RR interval distribution (histogram) including boxplot to visualize distribution and median
    * `bottom left`: Power Spectral Density (PSD) plot of RR intervals
    * `right`: Poincaré plot of RR intervals

    To use this function, either simply pass an :class:`~biopsykit.signals.ecg.EcgProcessor` object together with
    a ``key`` indicating which phase needs to be processed should be processed or the two dataframes ``ecg_signal``
    and ``heart_rate`` resulting from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.


    Parameters
    ----------
    ecg_processor : :class:`~biopsykit.signals.ecg.EcgProcessor`, optional
        ``EcgProcessor`` object. If this argument is passed, the ``key`` argument needs to be supplied as well.
    key : str, optional
        Dictionary key of the phase to process. Needed when ``ecg_processor`` is passed as argument.
    ecg_signal : :class:`~pandas.DataFrame`, optional
        dataframe with processed ECG signal. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
    rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`, optional
        Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
    sampling_rate : float, optional
        Sampling rate of recorded data in Hz. Not needed if ``ecg_processor`` is supplied as parameter.
        Default: 256
    plot_psd : bool, optional
        Whether to plot power spectral density (PDF) from frequency-based HRV analysis in a subplot or not.
        Default: ``True``
    **kwargs
        Additional parameters to configure the plot. Parameters include:

        * ``figsize``: Figure size
        * ``title``: Optional name to add to plot title (after "Electrocardiogram (ECG)")
        * ``legend_loc``: Location of legend in plot. Passed as `loc` parameter to :meth:`matplotlib.axes.Axes.legend`.
        * ``legend_fontsize``: Fontsize of legend text. Passed as `fontsize` parameter to
          :meth:`matplotlib.axes.Axes.legend`.
        * ``ax``: Pre-existing axes for the plot. Otherwise, a new figure and axes object are created and returned.

    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object


    See Also
    --------
    :func:`~biopsykit.signals.ecg.plotting.rr_distribution_plot`
        plot RR interval distribution
    :func:`~biopsykit.signals.ecg.plotting.hrv_poincare_plot`
        plot HRV using Poincaré plot
    :func:`~biopsykit.signals.ecg.plotting.hrv_frequency_plot`
        plot Power Spectral Density (PSD) of RR intervals

    """
    from biopsykit.signals.ecg import EcgProcessor  # pylint:disable=import-outside-toplevel

    _assert_ecg_input(ecg_processor, key, ecg_signal, rpeaks)

    title = kwargs.get("title", None)
    plt.rcParams["mathtext.default"] = "regular"

    if ecg_processor is not None:
        rpeaks = ecg_processor.rpeaks[key]
        sampling_rate = ecg_processor.sampling_rate

    # perform R peak correction before computing HRV measures
    rpeaks = EcgProcessor.correct_rpeaks(rpeaks=rpeaks, sampling_rate=sampling_rate)

    fig = plt.figure(constrained_layout=False, figsize=kwargs.get("figsize", (14, 7)))

    if plot_psd:
        spec = gs.GridSpec(ncols=2, nrows=2, height_ratios=[1, 1], width_ratios=[1, 1])
        axs = {
            "dist": fig.add_subplot(spec[0, :-1]),
            "freq": fig.add_subplot(spec[1, :-1]),
        }
    else:
        spec = gs.GridSpec(ncols=2, nrows=1, width_ratios=[1, 1])
        axs = {"dist": fig.add_subplot(spec[0, :-1])}

    spec_within = gs.GridSpecFromSubplotSpec(4, 4, subplot_spec=spec[:, -1], wspace=0.025, hspace=0.05)
    axs["poin"] = fig.add_subplot(spec_within[1:4, 0:3])
    axs["poin_x"] = fig.add_subplot(spec_within[0, 0:3])
    axs["poin_y"] = fig.add_subplot(spec_within[1:4, 3])
    axs["poin"].get_shared_x_axes().join(axs["poin"], axs["poin_x"])
    axs["poin"].get_shared_y_axes().join(axs["poin"], axs["poin_y"])

    if title:
        fig.suptitle(f"Heart Rate Variability (HRV) - {title}", fontweight="bold")
    else:
        fig.suptitle("Heart Rate Variability (HRV)", fontweight="bold")

    rr_distribution_plot(rpeaks, sampling_rate, ax=axs["dist"])
    hrv_poincare_plot(rpeaks, sampling_rate, axs=[axs["poin"], axs["poin_x"], axs["poin_y"]])
    if plot_psd:
        hrv_frequency_plot(rpeaks, sampling_rate, ax=axs["freq"])

    fig.tight_layout()
    return fig, list(axs.values())


def hr_distribution_plot(heart_rate: pd.DataFrame, **kwargs) -> Tuple[plt.Figure, plt.Axes]:
    """Plot heart rate distribution (histogram).

    This plot is also used as subplot in :func:`~biopsykit.signals.ecg.plotting.ecg_plot`.


    Parameters
    ----------
    heart_rate : :class:`~pandas.DataFrame`, optional
        dataframe with heart rate output. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`
    **kwargs
        Additional parameters to configure the plot. Parameters include:

        * ``figsize``: Figure size
        * ``ax``: Pre-existing axes for the plot. Otherwise, a new figure and axes object are created and returned.


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object

    See Also
    --------
    :func:`~biopsykit.signals.ecg.plotting.ecg_plot`
        plot ECG overview

    """
    ax: plt.Axes = kwargs.get("ax", None)
    plt.rcParams["mathtext.default"] = "regular"

    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    else:
        fig = ax.get_figure()

    ax = sns.histplot(heart_rate, color=colors_all.tech, ax=ax, kde=True, legend=False)

    ax.set_title("Heart Rate Distribution")
    ax.set_xlabel("Heart Rate [bpm]")
    ax.set_xlim(heart_rate.min().min() - 1, heart_rate.max().max() + 1)
    ax.tick_params(axis="x", which="major", bottom=True, labelbottom=True)

    ax.set_yticks([])
    ax.set_ylabel("")

    fig.tight_layout()
    return fig, ax


def rr_distribution_plot(
    rpeaks: pd.DataFrame, sampling_rate: Optional[int] = 256, **kwargs
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot distribution of RR intervals (histogram) with boxplot and rugplot.

    This plot is also used as subplot in :func:`~biopsykit.signals.ecg.plotting.hrv_plot`.


    Parameters
    ----------
    rpeaks : :class:`~pandas.DataFrame`, optional
        dataframe with R peaks. Output of :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
    sampling_rate : float, optional
        Sampling rate of recorded data in Hz. Default: 256
    **kwargs
        Additional parameters to configure the plot. Parameters include:

        * ``figsize``: Figure size
        * ``ax``: Pre-existing axes for the plot. Otherwise, a new figure and axes object are created and returned.

    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object

    See Also
    --------
    :func:`~biopsykit.signals.ecg.plotting.hr_distribution_plot`
        plot heart rate distribution (without boxplot and rugplot)

    """
    ax: plt.Axes = kwargs.get("ax", None)
    plt.rcParams["mathtext.default"] = "regular"

    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    else:
        fig = ax.get_figure()

    rri = _get_rr_intervals(rpeaks, sampling_rate)

    sns.set_palette(sns.light_palette(colors_all.fau, 3, reverse=True)[:-1])
    sns.histplot(rri, ax=ax, bins=10, kde=False, alpha=0.5, zorder=1)
    sns.rugplot(rri, ax=ax, lw=1.5, height=0.05, zorder=2)
    ax2 = ax.twinx()
    sns.kdeplot(rri, ax=ax2, lw=2.0, zorder=1)
    ax2.set_ylim(0)
    ax2.axis("off")

    ax.tick_params(axis="both", left=True, bottom=True)
    ax.set_title("Distribution of RR Intervals")
    ax.set_xlabel("RR Intervals [ms]")
    ax.set_ylabel("Count")

    tech_light = sns.light_palette(colors_all.tech, 3)[1]
    ax2.boxplot(
        rri,
        vert=False,
        positions=[ax2.get_ylim()[-1] / 10 + 0.05 * ax2.get_ylim()[-1]],
        widths=ax2.get_ylim()[-1] / 10,
        manage_ticks=False,
        patch_artist=True,
        boxprops={
            "linewidth": 2.0,
            "color": tech_light,
            "facecolor": colors_all.tech,
        },
        medianprops={"linewidth": 2.0, "color": tech_light},
        whiskerprops={"linewidth": 2.0, "color": tech_light},
        capprops={"linewidth": 2.0, "color": tech_light},
        zorder=4,
    )

    ax.set_xlim(0.95 * np.min(rri), 1.05 * np.max(rri))

    fig.tight_layout()
    return fig, ax


def individual_beats_plot(
    ecg_signal: pd.DataFrame, rpeaks: Optional[Sequence[int]] = None, sampling_rate: Optional[int] = 256, **kwargs
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot all segmented heart beats overlaid on top of each other.

    This plot is also used as subplot in :func:`~biopsykit.signals.ecg.plotting.ecg_plot`.


    Parameters
    ----------
    ecg_signal : :class:`~biopsykit.utils.datatype_helper.EcgResultDataFrame`, optional
        Dataframe with processed ECG signal. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
    rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`, optional
        Dataframe with detected R peaks or ``None`` to infer R peaks from ``ecg_signal``. Default: ``None``
    sampling_rate : float, optional
        Sampling rate of recorded data in Hz. Default: 256
    **kwargs
        Additional parameters to configure the plot. Parameters include:

        * ``figsize``: Figure size
        * ``ax``: Pre-existing axes for the plot. Otherwise, a new figure and axes object are created and returned.


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object


    See Also
    --------
    :func:`~biopsykit.signals.ecg.plotting.ecg_plot`
        plot ECG overview

    """
    ax: plt.Axes = kwargs.get("ax", None)
    plt.rcParams["mathtext.default"] = "regular"

    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    else:
        fig = ax.get_figure()

    if rpeaks is None:
        rpeaks = np.where(ecg_signal["ECG_R_Peaks"] == 1)[0]

    heartbeats = nk.ecg_segment(ecg_signal["ECG_Clean"], rpeaks, sampling_rate)
    heartbeats = nk.epochs_to_df(heartbeats)
    heartbeats_pivoted = heartbeats.pivot(index="Time", columns="Label", values="Signal")

    ax.set_title("Individual Heart Beats")
    ax.margins(x=0)

    # Aesthetics of heart beats
    cmap = iter(plt.cm.YlOrRd(np.linspace(0, 1, num=int(heartbeats["Label"].nunique()))))

    for x, color in zip(heartbeats_pivoted, cmap):
        ax.plot(heartbeats_pivoted[x], color=color)

    ax.set_yticks([])
    ax.tick_params(axis="x", which="major", bottom=True, labelbottom=True)

    fig.tight_layout()
    return fig, ax


# def ecg_plot_artifacts(ecg_signals: pd.DataFrame, sampling_rate: Optional[int] = 256):
#     # TODO not implemented yet
#     # Plot artifacts
#     _, rpeaks = nk.ecg_peaks(ecg_signals["ECG_Clean"], sampling_rate=sampling_rate)
#     _, _ = nk.signal_fixpeaks(rpeaks, sampling_rate=sampling_rate, iterative=True, show=True)


def hrv_poincare_plot(
    rpeaks: pd.DataFrame, sampling_rate: Optional[int] = 256, **kwargs
) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
    """Plot Heart Rate Variability as Poincaré Plot.

    This plot is also used as subplot in :func:`~biopsykit.signals.ecg.plotting.hrv_plot`.


    Parameters
    ----------
    rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`
            Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
    sampling_rate : float, optional
        Sampling rate of recorded data in Hz. Default: 256
    **kwargs
        Additional parameters to configure the plot. Parameters include:

        * ``figsize``: Figure size
        * ``ax``: List of pre-existing axes for the plot. Otherwise, a new figure and list of axes objects are
          created and returned.


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        Figure object
    axs : list of :class:`~matplotlib.axes.Axes`
        list of subplot axes objects

    See Also
    --------
    :func:`~biopsykit.signals.ecg.plotting.hrv_plot`
        plot heart rate variability

    """
    axs: List[plt.Axes] = kwargs.get("axs", None)
    plt.rcParams["mathtext.default"] = "regular"

    if axs is None:
        fig = plt.figure(figsize=kwargs.get("figsize", (8, 8)))
        axs = []
        # Prepare figure
        spec = gs.GridSpec(4, 4)
        axs.append(plt.subplot(spec[1:4, 0:3]))
        axs.append(plt.subplot(spec[0, 0:3]))
        axs.append(plt.subplot(spec[1:4, 3]))
        spec.update(wspace=0.025, hspace=0.05)  # Reduce spaces
    else:
        fig = axs[0].get_figure()

    rri = _get_rr_intervals(rpeaks, sampling_rate)

    mean_rri = float(np.mean(rri))
    sd = np.ediff1d(rri)
    sdsd = np.std(sd, ddof=1)
    sd1 = sdsd / np.sqrt(2)
    sd2 = np.sqrt(2 * np.std(rri, ddof=1) ** 2 - sd1**2)

    area = np.pi * sd1 * sd2

    sns.set_palette(sns.light_palette(colors_all.fau, 3, reverse=True)[:-1])
    sns.kdeplot(
        x=rri[:-1],
        y=rri[1:],
        ax=axs[0],
        n_levels=20,
        fill=True,
        thresh=0.05,
        alpha=0.8,
    )
    sns.scatterplot(x=rri[:-1], y=rri[1:], ax=axs[0], alpha=0.5, edgecolor=colors_all.fau)
    sns.histplot(x=rri[:-1], bins=int(len(rri) / 10), ax=axs[1], edgecolor="none")
    sns.histplot(y=rri[1:], bins=int(len(rri) / 10), ax=axs[2], edgecolor="none")

    ellipse = mpl.patches.Ellipse(
        (mean_rri, mean_rri),
        width=2 * sd2,
        height=2 * sd1,
        angle=45,
        ec=colors_all.fau,
        fc=colors_all.fau_dark,
        alpha=0.8,
    )
    axs[0].add_artist(ellipse)

    na = 4
    arr_sd1 = axs[0].arrow(
        mean_rri,
        mean_rri,
        -(sd1 - na) * np.cos(np.deg2rad(45)),
        (sd1 - na) * np.cos(np.deg2rad(45)),
        head_width=na,
        head_length=na,
        linewidth=2.0,
        ec=colors_all.phil,
        fc=colors_all.phil,
        zorder=4,
    )
    arr_sd2 = axs[0].arrow(
        mean_rri,
        mean_rri,
        (sd2 - na) * np.cos(np.deg2rad(45)),
        (sd2 - na) * np.sin(np.deg2rad(45)),
        head_width=na,
        head_length=na,
        linewidth=2.0,
        ec=colors_all.med,
        fc=colors_all.med,
        zorder=4,
    )
    axs[0].add_line(
        mpl.lines.Line2D(
            (np.min(rri), np.max(rri)),
            (np.min(rri), np.max(rri)),
            c=colors_all.med,
            ls=":",
            lw=2.0,
            alpha=0.8,
        )
    )
    axs[0].add_line(
        mpl.lines.Line2D(
            (
                mean_rri - sd1 * np.cos(np.deg2rad(45)) * na,
                mean_rri + sd1 * np.cos(np.deg2rad(45)) * na,
            ),
            (
                mean_rri + sd1 * np.sin(np.deg2rad(45)) * na,
                mean_rri - sd1 * np.sin(np.deg2rad(45)) * na,
            ),
            c=colors_all.phil,
            ls=":",
            lw=2.0,
            alpha=0.8,
        )
    )
    # for Area and SD1/SD2 in Legend
    a3 = mpl.patches.Patch(facecolor="white", alpha=0.0)
    a4 = mpl.patches.Patch(facecolor="white", alpha=0.0)

    axs[0].legend(
        [arr_sd1, arr_sd2, a3, a4],
        [
            "SD1: $%.3f ms$" % sd1,
            "SD2: $%.3f ms$" % sd2,
            "S: $%.3f ms^2$" % area,
            "SD1/SD2: %.3f" % (sd1 / sd2),
        ],
        framealpha=1,
        fontsize="small",
    )

    axs[0].set_xlabel(r"$RR_{i}~[ms]$")
    axs[0].set_ylabel(r"$RR_{i+1}~[ms]$")
    axs[0].xaxis.set_major_locator(mticks.MultipleLocator(100))
    axs[0].xaxis.set_minor_locator(mticks.MultipleLocator(25))
    axs[0].yaxis.set_major_locator(mticks.MultipleLocator(100))
    axs[0].yaxis.set_minor_locator(mticks.MultipleLocator(25))
    axs[0].tick_params(axis="both", which="both", left=True, bottom=True)
    axs[0].tick_params(axis="x", labelrotation=30)
    axs[1].axis("off")
    axs[2].axis("off")

    fig.tight_layout()
    return fig, axs


def hrv_frequency_plot(
    rpeaks: pd.DataFrame, sampling_rate: Optional[int] = 256, **kwargs
) -> Tuple[plt.Figure, plt.Axes]:
    """Plot Power Spectral Density (PSD) of RR intervals.

    This plot is also used as subplot in :func:`~biopsykit.signals.ecg.plotting.hrv_plot`.


    Parameters
    ----------
    rpeaks : :class:`~biopsykit.utils.datatype_helper.RPeakDataFrame`
            Dataframe with detected R peaks. Output from :meth:`~biopsykit.signals.ecg.EcgProcessor.ecg_process()`.
    sampling_rate : float, optional
        Sampling rate of recorded data in Hz. Default: 256
    **kwargs
        Additional parameters to configure the plot. Parameters include:

        * ``figsize``: Figure size
        * ``ax``: Pre-existing axes for the plot. Otherwise, a new figure and axes object are created and returned.


    Returns
    -------
    fig : :class:`~matplotlib.figure.Figure`
        figure object
    ax : :class:`~matplotlib.axes.Axes`
        axes object


    See Also
    --------
    :func:`~biopsykit.signals.ecg.plotting.hrv_plot`
        plot heart rate variability

    """
    ax: plt.Axes = kwargs.get("ax", None)
    plt.rcParams["mathtext.default"] = "regular"

    if ax is None:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize"))
    else:
        fig = ax.get_figure()

    rpeaks = sanitize_input_1d(rpeaks["R_Peak_Idx"])

    rri, rri_time, _ = _hrv_get_rri(rpeaks, sampling_rate=sampling_rate)
    rri, rri_time, fs = nk.intervals_process(
        rri, intervals_time=rri_time, interpolate=True, interpolation_rate=sampling_rate
    )
    hrv = nk.hrv_frequency(rpeaks, sampling_rate)
    out_bands = hrv[["HRV_ULF", "HRV_VLF", "HRV_LF", "HRV_HF", "HRV_VHF"]]
    out_bands.columns = [col.replace("HRV_", "") for col in out_bands.columns]
    _hrv_frequency_show(rri, out_bands, sampling_rate=fs, ax=ax)

    ax.set_title("Power Spectral Density (PSD)")
    ax.set_ylabel("Spectrum $[{ms}^2/Hz]$")
    ax.set_xlabel("Frequency [Hz]")

    ax.tick_params(axis="both", left=True, bottom=True)
    ax.margins(x=0)
    ax.set_ylim(0)

    fig.tight_layout()
    return fig, ax


def _get_rr_intervals(rpeaks: pd.DataFrame, sampling_rate: Optional[int] = 256) -> np.array:
    rri = (np.ediff1d(rpeaks["R_Peak_Idx"], to_begin=0) / sampling_rate) * 1000
    rri = rri[1:]
    return rri



"""Module for processing IMU data."""
from biopsykit.signals.imu import (
    activity_counts,
    feature_extraction,
    rest_periods,
    static_moment_detection,
    wear_detection,
)
from biopsykit.signals.imu.imu import convert_acc_data_to_g, sliding_windows_imu, var_norm_windows

__all__ = [
    "var_norm_windows",
    "convert_acc_data_to_g",
    "sliding_windows_imu",
    "rest_periods",
    "wear_detection",
    "static_moment_detection",
    "activity_counts",
    "feature_extraction",
]



"""A Python package for the analysis of biopsychological data."""

from biopsykit import (
    classification,
    example_data,
    io,
    metadata,
    plotting,
    protocols,
    questionnaires,
    saliva,
    signals,
    sleep,
    stats,
    utils,
)

__all__ = [
    "classification",
    "example_data",
    "io",
    "metadata",
    "plotting",
    "protocols",
    "questionnaires",
    "saliva",
    "signals",
    "sleep",
    "stats",
    "utils",
]

__version__ = "0.10.0"


def version() -> None:
    """Get the version of BioPsyKit and its core dependencies.

    Examples
    --------
    >>> import biopsykit as bp
    >>>
    >>> bp.version()

    """
    import platform

    import matplotlib
    import neurokit2
    import numpy as np
    import pandas as pd
    import pingouin
    import scipy

    print(
        f"Operating System: {platform.system()} ({platform.architecture()[1]} {platform.architecture()[0]})\n",
        f"- Python: {platform.python_version()}\n",
        f"- BioPsyKit: {__version__}\n\n",
        f"- NumPy: {np.__version__}\n",
        f"- Pandas: {pd.__version__}\n",
        f"- SciPy: {scipy.__version__}\n",
        f"- matplotlib: {matplotlib.__version__}\n",
        f"- NeuroKit2: {neurokit2.__version__}\n",
        f"- pingouin: {pingouin.__version__}\n",
    )



"""Module for generating Activity Counts from raw acceleration signals."""
import datetime
from typing import Optional, Union

import numpy as np
import pandas as pd
import pytz
from biopsykit.utils._types import arr_t
from biopsykit.utils.array_handling import add_datetime_index, downsample, sanitize_input_nd
from biopsykit.utils.datatype_helper import is_acc1d_dataframe, is_acc3d_dataframe
from biopsykit.utils.time import tz
from scipy import signal


class ActivityCounts:
    """Generate Activity Counts from raw acceleration signals.

    Actigraph Activity Counts are a unit used in many human activity studies.
    However, it can only be outputted by the official Actigraph Software.
    The following implementation uses a reverse engineered version of the Actigraph filter based on
    (Brønd et al., 2017).


    References
    ----------
    Brønd, J. C., Andersen, L. B., & Arvidsson, D. (2017). Generating ActiGraph Counts from Raw Acceleration Recorded
    by an Alternative Monitor. *Medicine and Science in Sports and Exercise*, 49(11), 2351-2360.
    https://doi.org/10.1249/MSS.0000000000001344

    """

    data: pd.DataFrame = None
    sampling_rate: float = None
    activity_counts_: np.ndarray = None
    timezone: datetime.tzinfo = tz

    def __init__(self, sampling_rate: float, timezone: Optional[str] = None):
        """Initialize a new ``ActivityCounts`` instance.

        Parameters
        ----------
        sampling_rate : float
            sampling rate of recorded data in Hz
        timezone: str
            timezone to which wear times will be converted

        """
        self.sampling_rate = sampling_rate
        if timezone:
            self.timezone = pytz.timezone(timezone)

    @staticmethod
    def _compute_norm(data: np.ndarray) -> np.ndarray:
        return np.linalg.norm(data, axis=1)

    @staticmethod
    def _aliasing_filter(data: np.ndarray, sampling_rate: Union[int, float]) -> np.ndarray:
        sos = signal.butter(5, [0.01, 7], "bp", fs=sampling_rate, output="sos")
        return signal.sosfiltfilt(sos, data)

    @staticmethod
    def _actigraph_filter(data: np.ndarray) -> np.ndarray:
        b = [
            0.04910898,
            -0.12284184,
            0.14355788,
            -0.11269399,
            0.05380374,
            -0.02023027,
            0.00637785,
            0.01851254,
            -0.03815411,
            0.04872652,
            -0.05257721,
            0.04784714,
            -0.04601483,
            0.03628334,
            -0.01297681,
            -0.00462621,
            0.01283540,
            -0.00937622,
            0.00344850,
            -0.00080972,
            -0.00019623,
        ]
        a = [
            1.00000000,
            -4.16372603,
            7.57115309,
            -7.98046903,
            5.38501191,
            -2.46356271,
            0.89238142,
            0.06360999,
            -1.34810513,
            2.47338133,
            -2.92571736,
            2.92983230,
            -2.78159063,
            2.47767354,
            -1.68473849,
            0.46482863,
            0.46565289,
            -0.67311897,
            0.41620323,
            -0.13832322,
            0.01985172,
        ]
        return signal.filtfilt(b, a, data)

    @staticmethod
    def _downsample(
        data: np.ndarray,
        sampling_rate: Union[int, float],
        final_sampling_rate: Union[int, float],
    ) -> np.ndarray:
        return downsample(data, sampling_rate, final_sampling_rate)

    @staticmethod
    def _truncate(data: np.ndarray) -> np.ndarray:
        upper_threshold = 2.13  # g
        lower_threshold = 0.068  # g
        data[data > upper_threshold] = upper_threshold
        data[data < lower_threshold] = 0
        return data

    @staticmethod
    def _digitize_8bit(data: np.ndarray) -> np.ndarray:
        max_val = 2.13  # g
        data //= max_val / (2**7)
        return data

    @staticmethod
    def _accumulate_second_bins(data: np.ndarray) -> np.ndarray:
        n_samples = 10
        #  Pad data at end to "fill" last bin
        padded_data = np.pad(data, (0, n_samples - len(data) % n_samples), "constant", constant_values=0)
        return padded_data.reshape((len(padded_data) // n_samples, -1)).sum(axis=1)

    def calculate(self, data: arr_t) -> arr_t:
        """Calculate Activity Counts from acceleration data.

        Parameters
        ----------
        data : array_like
            input data. Must either be 3-d or 1-d (e.g., norm, or a specific axis) acceleration data

        Returns
        -------
        array_like
            output data with Activity Counts

        """
        start_idx = None
        if isinstance(data, pd.DataFrame):
            # if dataframe, assert to be a acceleration dataframe according to biopsykit's convention
            if data.shape[1] == 3:
                is_acc3d_dataframe(data)
            if data.shape[1] == 1:
                is_acc1d_dataframe(data)
            data = data.filter(like="acc")
            if isinstance(data.index, pd.DatetimeIndex):
                start_idx = data.index[0]

        arr = sanitize_input_nd(data, ncols=(1, 3))

        if arr.shape[1] not in (1, 3):
            raise ValueError(
                f"{self.__class__.__name__} takes only 1-d or 3-d accelerometer data! Got {arr.shape[1]}-d data."
            )
        if arr.shape[1] != 1:
            arr = self._compute_norm(arr)

        arr = self._downsample(arr, self.sampling_rate, 30)
        arr = self._aliasing_filter(arr, 30)
        arr = self._actigraph_filter(arr)
        arr = self._downsample(arr, 30, 10)
        arr = np.abs(arr)
        arr = self._truncate(arr)
        arr = self._digitize_8bit(arr)
        arr = self._accumulate_second_bins(arr)
        if start_idx is not None:
            arr = add_datetime_index(arr, start_idx, 1 / 60, column_name=["activity_counts"])

        return arr



"""Module with utility functions to handle sleep data."""
import datetime
from typing import Dict

import numpy as np
import pandas as pd
from biopsykit.utils._datatype_validation_helper import _assert_is_dtype


def split_nights(data: pd.DataFrame) -> Dict[str, pd.DataFrame]:
    """Split continuous data into individual nights.

    This function splits data into individual nights. The split is performed at 6pm because that's the time of day
    where the probability of sleeping is the lowest.

    Parameters
    ----------
    data : :class:`~pandas.DataFrame`
        input data

    Returns
    -------
    dict
        dictionary with recording date (keys) and dataframes split into individual nights (values).
        By convention, if the recording of one night started after 12 am the recording date is set to the previous date.

    """
    _assert_is_dtype(data.index, pd.DatetimeIndex)
    # split data per day
    date_diff = np.diff(data.index.date)
    date_diff = np.append(date_diff[0], date_diff)
    idx_date = np.where(date_diff)[0]

    # split data per time: split at 6 pm, because that's the time of day where the probability of sleeping is the lowest
    time_diff = data.index.time <= pd.Timestamp("18:00:00").time()
    time_diff = np.append(time_diff[0], time_diff)
    idx_time = np.where(np.diff(time_diff))[0]

    # concatenate both splitting criteria and split data
    idx_split = np.unique(np.concatenate([idx_date, idx_time]))
    data_split = np.split(data, idx_split)

    time_6pm = pd.Timestamp("18:00:00").time()

    # concatenate data from one night (data between 6 pm and 12 am from the previous day and
    # between 12 am and 6 pm of the next day)
    for i, df in enumerate(data_split):
        if i < (len(data_split) - 1):
            df_curr = df
            df_next = data_split[i + 1]

            date_curr = df_curr.index[0].date()
            date_next = df_next.index[0].date()
            time_curr = df_curr.index[0].time()
            time_next = df_next.index[0].time()

            # check if dates are consecutive and if first part is after 6 pm and second part is before 6 am
            if (date_next == date_curr + datetime.timedelta(days=1)) and (time_curr > time_6pm > time_next):
                data_split[i] = pd.concat([df_curr, df_next])
                # delete the second part
                del data_split[i + 1]

    # create dict with data from each night. dictionary keys are the dates.
    dict_data = {}
    for df in data_split:
        date = df.index[0].normalize().date()
        # By convention, if the recording started after 12 am the recording date is set to the previous date
        if df.index[0].time() < pd.Timestamp("18:00:00").time():
            date = date - pd.Timedelta("1d")
        dict_data[str(date)] = df

    return dict_data
